import click

from .. import controllers
from ..main import cli


@cli.command(name='package')
@click.argument('id_or_url')
def _package(id_or_url: str):
    """Package the specified novel to epub"""
    controllers.package(id_or_url)


@cli.command(name='process')
@click.argument('id_or_url')
@click.option('--limit', type=int, help='Maximum count of chapters to update. Set to 0 or less to skip.')
@click.option('--browser', help='Extract cookies from the specified browser and use them in subsequent requests.')
@click.option('--threads', type=int, help="Amount of threads to use when downloading chapters.")
def _process(id_or_url: str, limit: int, browser: str, threads: int):
    """Runs 'update' and 'package' commands consecutively"""
    controllers.update(id_or_url, browser, limit, threads)
    controllers.package(id_or_url)


@cli.command(name='update')
@click.argument('id_or_url')
@click.option('--limit', type=int, help='Maximum count of chapters to update. Set to 0 or less to skip.')
@click.option('--browser', help='Extract cookies from the specified browser and use them in subsequent requests.')
@click.option('--threads', type=int, help="Amount of threads to use when downloading chapters.")
def _update(id_or_url: str, limit: int, browser: str, threads: int):
    """Scrape the website of the novel and update the database"""
    controllers.update(id_or_url, browser, limit, threads)


@cli.command(name='metadata')
@click.argument('id_or_url')
@click.argument('metadata_url')
def _metadata(id_or_url: str, metadata_url):
    """Import metadata from a novel metadata provider"""
    controllers.import_metadata(id_or_url, metadata_url)


@cli.command(name='list')
def _list():
    """List all novels saved in database"""
    controllers.list_novels()


@cli.command(name='info')
@click.argument('id_or_url')
def _info(id_or_url: str):
    """Show saved information of a novel"""
    controllers.show_info(id_or_url)


@cli.group(name='novel')
def _novel():
    """Manage more specific aspects of novels"""


@_novel.command(name='clean')
@click.argument('id_or_url')
@click.option('--content-only', is_flag=True, help="Only remove chapter content")
def _clean_novel(id_or_url: str, content_only: bool):
    """Removes all except vital information related to novel"""
    controllers.clean_novel(id_or_url, content_only)


@_novel.command(name='delete')
@click.argument('id_or_url')
def _delete_novel(id_or_url: str):
    """Remove novel entry from database"""
    controllers.delete_novel(id_or_url)


@_novel.group(name='url')
def _url():
    """Handles novel url operations, add or delete"""


@_url.command(name='add')
@click.argument('id_or_url')
@click.argument('new_url')
def _add_url(id_or_url: str, new_url: str):
    """Add a url from a select novel"""
    controllers.add_url(id_or_url, new_url)


@_url.command(name='remove')
@click.argument('url')
def _remove_url(url: str):
    """Remove a url from a select novel"""
    controllers.remove_url(url)



from pathlib import Path

from novelsave.core.services import BaseFileService


class FileService(BaseFileService):

    def write_str(self, path: Path, data: str):
        """write and replace strings to a file"""
        with path.open('w') as file:
            file.write(data)

    def write_bytes(self, path: Path, data: bytes):
        """write and replace bytes to a file"""
        with path.open('wb') as file:
            file.write(data)

    def read_str(self, path: Path) -> str:
        """read strings from a file"""
        with path.open('r') as file:
            return file.read()

    def read_bytes(self, path: Path) -> bytes:
        """read bytes from a file"""
        with path.open('rb') as file:
            return file.read()



from abc import ABC, abstractmethod


class BaseMetaService(ABC):

    @property
    @abstractmethod
    def current_version(self) -> str:
        """Current application version"""

    @abstractmethod
    def get_latest_version(self) -> str:
        """Latest stable version of novelsave available for download"""



from pathlib import Path
from typing import Dict, Union
from urllib.parse import urlparse

from novelsave.core.entities.novel import Novel, Asset
from novelsave.core.services import BasePathService, BaseNovelService
from novelsave.core.services.source import BaseSourceService
from novelsave.exceptions import SourceNotFoundException
from novelsave.utils.helpers import string_helper


class PathService(BasePathService):

    def __init__(
            self,
            data_dir: Path,
            novels_dir: Path,
            division_rules: Dict[str, str],
            novel_service: BaseNovelService,
            source_service: BaseSourceService,
    ):
        self.data_dir = data_dir
        self.novels_dir = novels_dir
        self.division_rules = division_rules
        self.novel_service = novel_service
        self.source_service = source_service

    def divide(self, path: Path) -> Path:
        parent = path.parent / self.division_rules.get(path.suffix, '')
        return parent.resolve() / path.name

    def novel_save_path(self, novel: Novel) -> Path:
        url = self.novel_service.get_primary_url(novel)
        try:
            source_gateway = self.source_service.source_from_url(url)
            source_folder_name = source_gateway.name
        except SourceNotFoundException:
            source_folder_name = ''

        novel_name_slug = string_helper.slugify(novel.title, "_")

        return (self.novels_dir / source_folder_name / novel_name_slug).resolve()

    def novel_data_path(self, novel: Novel) -> Path:
        return self.data_dir / str(novel.id)

    def asset_path(self, novel: Novel, asset: Asset) -> Path:
        parse_result = urlparse(novel.thumbnail_url)
        suffix = Path(parse_result.path).suffix
        file = Path(str(asset.id) + suffix)

        return self.novel_data_path(novel) / self.divide(file)

    def thumbnail_path(self, novel: Novel) -> Path:
        suffix = Path(novel.thumbnail_url).suffix

        return self.data_dir / str(novel.id) / f'cover{suffix}'

    def resolve_data_path(self, r_path: Union[Path, str]) -> Path:
        return self.data_dir / str(r_path).lstrip('.').lstrip('/\\')

    def relative_to_novel_dir(self, path: Path) -> Path:
        return path.relative_to(self.novels_dir)

    def relative_to_data_dir(self, path: Path) -> Path:
        return path.relative_to(self.data_dir)



from pathlib import Path

from alembic.command import upgrade
from alembic.config import Config


def make_config(dir_: Path, url_: str, config_='alembic.ini'):
    """
    :param dir_: migrations script directory
    :param url_: sqlalchemy database url
    :param config_: config
    :return:
    """
    # retrieves config file path
    config_file = dir_ / config_

    config = Config(file_=config_file)
    config.set_main_option('script_location', str(dir_))
    config.set_main_option('sqlalchemy.url', url_)

    config.attributes['configure_logger'] = False

    return config


def migrate(url: str):
    config = make_config(Path(__file__).parent, url, 'alembic.ini')

    # upgrade the database to the latest revision
    upgrade(config, 'head')



from sqlalchemy import Column, Integer, String, TIMESTAMP, ForeignKey, func, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from ..base import Base


class Chapter(Base):
    __tablename__ = 'chapters'
    __table_args__ = (
        UniqueConstraint('volume_id', 'index'),
    )

    id = Column(Integer, primary_key=True)
    index = Column(Integer, nullable=False)
    title = Column(String, nullable=False)
    url = Column(String, nullable=False)

    content = Column(String, nullable=True)
    volume_id = Column(Integer, ForeignKey('volumes.id'), nullable=False)
    volume = relationship('Volume', back_populates='chapters')

    last_updated = Column(TIMESTAMP, server_default=func.now(), onupdate=func.current_timestamp())



from collections import defaultdict
from typing import Dict, List

from bs4 import BeautifulSoup
from loguru import logger
from sqlalchemy import update, delete
from sqlalchemy.orm import Session

from novelsave.core.dtos import ChapterDTO
from novelsave.core.entities.constants import AssetTypes
from novelsave.core.entities.novel import Asset, Novel
from novelsave.core.services import BaseAssetService, BasePathService
from novelsave.utils.helpers import url_helper


class AssetService(BaseAssetService):

    def __init__(
            self,
            session: Session,
            path_service: BasePathService,
    ):
        self.session = session
        self.path_service = path_service

    def downloaded_assets(self, novel: Novel) -> List[Asset]:
        downloaded = []
        for asset in novel.assets:
            if asset.path is None:
                continue

            file = self.path_service.resolve_data_path(asset.path)
            if not file.exists() or not file.is_file():
                continue

            downloaded.append(asset)

        return downloaded

    def pending_assets(self, novel: Novel) -> List[Asset]:
        pending = []
        for asset in novel.assets:
            if asset.path is None:
                pending.append(asset)
                continue

            file = self.path_service.resolve_data_path(asset.path)
            if not file.exists() or not file.is_file():
                pending.append(asset)

        return pending

    def update_asset_path(self, asset: Asset):
        self.session.execute(update(Asset).where(Asset.id == asset.id).values(path=asset.path))
        self.session.commit()

    def delete_assets_of_novel(self, novel: Novel):
        self.session.execute(delete(Asset).where(Asset.novel_id == novel.id))
        self.session.commit()

    def update_assets(self, novel: Novel, assets: List[Asset]) -> Dict[str, Asset]:
        indexed_assets = {a.url: a for a in novel.assets}
        indexed_specific = {}

        assets_to_add = []
        for asset in assets:
            try:
                indexed_specific[asset.url] = indexed_assets[asset.url]
            except KeyError:
                indexed_specific[asset.url] = asset
                assets_to_add.append(asset)

        # we only interact with the database if we have something to add
        # this is [expected] to be more performant
        if assets_to_add:
            logger.debug(f'Adding newly found assets (count={len(assets_to_add)}).')
            self.session.add_all(assets_to_add)
            self.session.flush()
        else:
            logger.debug(f"Skipping adding assets ({novel.id=}, reason='No assets to add')")

        return indexed_specific

    def collect_assets(self, novel: Novel, chapter: ChapterDTO) -> str:

        # using default parser since lxml inserts <html> and <body> tags
        # those would have to be removed since the input doesnt require to have them
        # so its better to not insert them at all
        soup = BeautifulSoup(chapter.content, 'html.parser')

        assets = []
        for img in soup.select('img'):
            src = img.get('src', default=None)
            alt = img.get('alt', default='[Unspecified]')
            if not src:
                continue

            url = url_helper.absolute_url(src, chapter.url)
            asset = Asset(
                name=alt,
                url=url,
                type_id=AssetTypes.IMAGE,
                novel_id=novel.id,
            )

            assets.append(asset)

        logger.debug(f"Identified asset images (novel='{novel.title}', count={len(assets)}).")
        if not assets:
            logger.debug(f"Skipped further asset processing (novel='{novel.title}', reason='No assets identified')")
            return chapter.content

        indexed_assets = self.update_assets(novel, assets)
        for img in soup.select('img'):
            src = img.get('src', default=None)
            if not src:
                continue

            url = url_helper.absolute_url(src, chapter.url)
            img['src'] = f'{{id{indexed_assets[url].id}}}'

        logger.debug(f"Embedded asset markers (chapter='{chapter.title}').")

        return str(soup)

    def mapping_dict(self, path_mapping: Dict[int, str]):
        return defaultdict(
            str,
            {f'id{key}': path for key, path in path_mapping.items()}
        )

    def inject_assets(self, html: str, mapping_dict: Dict[int, str]):
        return html.format_map(mapping_dict)



import sys

from dependency_injector.wiring import inject, Provide
from loguru import logger

from novelsave.cli.helpers import get_novel
from novelsave.containers import Application
from novelsave.core.services import BaseNovelService


@inject
def remove_url(
        url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
):
    """Removes the selected url from the database"""
    try:
        novel = get_novel(url)
    except ValueError:
        sys.exit(1)

    try:
        novel_service.remove_url(novel, url)
    except ValueError as e:
        logger.error(e)
        sys.exit(1)
    else:
        logger.info(f"Removed url from novel (id={novel.id}, title='{novel.title}', {url=}).")


@inject
def add_url(
        id_or_url: str,
        new_url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
):
    """Deletes the novel and all its data"""
    try:
        novel = get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    try:
        novel_service.add_url(novel, new_url.rstrip('/'))
    except ValueError as e:
        logger.error(e)
        sys.exit(1)
    else:
        logger.info(f"Url added to novel (id={novel.id}, title='{novel.title}', {new_url=}).")



"""initial

Revision ID: e5c4fb5600ea
Revises: 
Create Date: 2021-08-24 11:30:55.605615

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e5c4fb5600ea'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    asset_types = op.create_table('asset_types',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=16), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('novels',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('author', sa.String(), nullable=True),
    sa.Column('synopsis', sa.String(), nullable=True),
    sa.Column('lang', sa.String(), nullable=True),
    sa.Column('thumbnail_url', sa.String(), nullable=True),
    sa.Column('thumbnail_path', sa.String(), nullable=True),
    sa.Column('last_updated', sa.TIMESTAMP(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('assets',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('path', sa.String(), nullable=True),
    sa.Column('type_id', sa.Integer(), nullable=False),
    sa.Column('novel_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['novel_id'], ['novels.id'], ),
    sa.ForeignKeyConstraint(['type_id'], ['asset_types.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('novel_id', 'url')
    )
    op.create_table('novel_metadata',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('value', sa.String(), nullable=False),
    sa.Column('namespace', sa.String(), nullable=True),
    sa.Column('others', sa.String(), nullable=True),
    sa.Column('novel_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['novel_id'], ['novels.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('novel_urls',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('url', sa.String(), nullable=False),
    sa.Column('novel_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['novel_id'], ['novels.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('volumes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('index', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('novel_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['novel_id'], ['novels.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('novel_id', 'index')
    )
    op.create_table('chapters',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('index', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('url', sa.String(), nullable=False),
    sa.Column('content', sa.String(), nullable=True),
    sa.Column('volume_id', sa.Integer(), nullable=False),
    sa.Column('last_updated', sa.TIMESTAMP(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.ForeignKeyConstraint(['volume_id'], ['volumes.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('volume_id', 'index')
    )
    # ### end Alembic commands ###

    op.bulk_insert(
        asset_types,
        [{'id': 1, 'name': 'image'}]
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('chapters')
    op.drop_table('volumes')
    op.drop_table('novel_urls')
    op.drop_table('novel_metadata')
    op.drop_table('assets')
    op.drop_table('novels')
    op.drop_table('asset_types')
    # ### end Alembic commands ###



from abc import ABC, abstractmethod
from typing import List

from novelsave.core.dtos import MetaDataDTO


class BaseMetaSourceGateway(ABC):

    @property
    @abstractmethod
    def name(self):
        """name of the corresponding metadata source"""

    @abstractmethod
    def metadata_by_url(self, url: str) -> List[MetaDataDTO]:
        """get metadata from the url"""



import shutil
import sys

from dependency_injector.wiring import inject, Provide
from loguru import logger

from novelsave.cli import helpers as cli_helpers
from novelsave.containers import Application
from novelsave.core.services import BaseNovelService, BasePathService, BaseAssetService
from novelsave.core.services.source import BaseSourceService
from novelsave.exceptions import SourceNotFoundException


def show_info(id_or_url: str):
    """print current information of novel"""
    try:
        novel = cli_helpers.get_novel(id_or_url, silent=True)
    except ValueError:
        sys.exit(1)

    logger.info('[novel]')
    logger.info(f'id = {novel.id}')
    logger.info(f'title = {novel.title}')
    logger.info(f'author = {novel.author}')
    logger.info(f'lang = {novel.lang}')
    logger.info(f'thumbnail = {novel.thumbnail_url}')
    logger.info(f'synopsis = ')
    for line in novel.synopsis.splitlines():
        logger.info(f'{"":<4}{line.strip()}')


@inject
def list_novels(
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        source_service: BaseSourceService = Provide[Application.services.source_service],
):
    novels = novel_service.get_all_novels()
    for novel in novels:
        url = novel_service.get_primary_url(novel)

        try:
            source = source_service.source_from_url(url).name
        except SourceNotFoundException:
            source = None

        logger.info(f"Novel (id={novel.id}, title='{novel.title}', {source=}, last_updated='{novel.last_updated}').")


@inject
def delete_downloaded_content(
        id_or_url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
):
    """deletes all the downloaded content from chapters of novel"""
    try:
        novel = cli_helpers.get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    novel_service.delete_content(novel)
    logger.info(f"Deleted chapter content from novel (id={novel.id}, title='{novel.title}').")


@inject
def delete_associations(
        id_or_url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        path_service: BasePathService = Provide[Application.services.path_service],
        asset_service: BaseAssetService = Provide[Application.services.asset_service],
):
    """Removes all except vital information related to novel, this includes chapters, metadata, and assets."""
    try:
        novel = cli_helpers.get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    novel_service.delete_volumes(novel)
    logger.info(f"Deleted volumes and chapters of novel (id={novel.id}, title='{novel.title}').")

    novel_service.delete_metadata(novel)
    logger.info(f"Deleted metadata of novel (id={novel.id}, title='{novel.title}').")

    asset_service.delete_assets_of_novel(novel)
    logger.info(f"Deleted asset entries of novel (id={novel.id}, title='{novel.title}').")

    novel_dir = path_service.novel_data_path(novel)
    if novel_dir.exists():
        shutil.rmtree(novel_dir)
    logger.info(f"Deleted data of novel (path='{path_service.resolve_data_path(novel_dir)}')")


def clean_novel(id_or_url: str, content_only: bool):
    """Remove accessory fields from novel

    Removes all except vital information related to novel, this includes chapters, metadata, and assets.

    If content_only flag is specified, only the chapter content is deleted, keeping the chapter entries as is.
    """
    if content_only:
        delete_downloaded_content(id_or_url)
    else:
        delete_associations(id_or_url)


@inject
def delete_novel(
        id_or_url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        path_service: BasePathService = Provide[Application.services.path_service],
):
    """delete all records of novel. this includes chapters, and assets"""
    try:
        novel = cli_helpers.get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    novel_dir = path_service.novel_data_path(novel)
    if novel_dir.exists():
        shutil.rmtree(novel_dir)
    logger.info(f"Deleted data of novel (path='{path_service.resolve_data_path(novel_dir)}')")

    novel_service.delete_novel(novel)
    logger.info(f"Deleted novel (id={novel.id}, title='{novel.title}')")


@inject
def import_metadata(
        id_or_url: str,
        metadata_url: str,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
):
    """import metadata from a metadata supplied into an existing novel"""
    try:
        novel = cli_helpers.get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    meta_source_gateway = cli_helpers.get_meta_source_gateway(metadata_url)
    metadata_dtos = meta_source_gateway.metadata_by_url(metadata_url)

    novel_service.update_metadata(novel, metadata_dtos)



from abc import ABC, abstractmethod


class BaseConfigService(ABC):

    @abstractmethod
    def get_novel_dir(self):
        """get novels save directory"""

    @abstractmethod
    def set_novel_dir(self, path):
        """set novels save directory"""

    @abstractmethod
    def reset_novel_dir(self):
        """reset novels save directory"""



from ._update import update
from ._package import package
from ._url import add_url, remove_url
from ._novel import show_info, list_novels, delete_downloaded_content, clean_novel, delete_novel, import_metadata



from pathlib import Path
from typing import Optional, Callable, List, Dict, Tuple

from loguru import logger
from sqlalchemy import delete, select, update
from sqlalchemy.orm import Session

from novelsave.core.dtos import NovelDTO, ChapterDTO, MetaDataDTO, VolumeDTO
from novelsave.core.entities.novel import Novel, NovelUrl, Chapter, Volume, MetaData
from novelsave.core.services import BaseNovelService
from novelsave.services import FileService
from novelsave.utils.adapters import DTOAdapter


class NovelService(BaseNovelService):

    def __init__(
            self,
            session: Session,
            dto_adapter: DTOAdapter,
            file_service: FileService,
    ):
        self.session = session
        self.dto_adapter = dto_adapter
        self.file_service = file_service

    def get_all_novels(self) -> List[Novel]:
        return self.session.execute(select(Novel)).scalars().all()

    def get_novel_by_id(self, id_: int) -> Novel:
        """retrieve a novel from persistence using its id"""
        return self.session.get(Novel, id_)

    def get_novel_by_url(self, url: str) -> Optional[Novel]:
        """retrieve a novel from persistence using its url"""
        return self.session.query(Novel).join(NovelUrl).filter(NovelUrl.url == url).first()

    def get_primary_url(self, novel: Novel) -> str:
        return novel.urls[0].url

    def get_urls(self, novel) -> List[NovelUrl]:
        return novel.urls

    def get_chapters(self, novel: Novel) -> List[Chapter]:
        return self.session.execute(
            select(Chapter).join(Volume).where(Volume.novel_id == novel.id)
        ).scalars().all()

    def get_pending_chapters(self, novel: Novel, limit: int = -1):
        stmt = select(Chapter).join(Volume).where((Chapter.content == None) & (Volume.novel_id == novel.id))
        if limit is not None and limit > 0:
            stmt = stmt.limit(limit)

        return self.session.execute(stmt).scalars().all()

    def get_volumes(self, novel: Novel) -> List[Volume]:
        return novel.volumes

    def get_volumes_with_chapters(self, novel: Novel) -> Dict[Volume, List[Chapter]]:
        volumes = self.session.execute(select(Volume).where(Volume.novel_id == novel.id)).scalars().all()

        return {
            volume: self.session.execute(
                select(Chapter).where((Chapter.content != None) & (Chapter.volume_id == volume.id))
            ).scalars().all()
            for volume in volumes
        }

    def get_metadata(self, novel: Novel) -> List[MetaData]:
        return novel.novel_metadata

    def insert_novel(self, novel_dto: NovelDTO) -> Novel:
        novel, url = self.dto_adapter.novel_from_dto(novel_dto)
        self.session.add(novel)
        self.session.flush()

        url.novel_id = novel.id
        self.session.add(url)
        self.session.commit()

        return novel

    def insert_chapters(
            self,
            novel: Novel,
            volume_dtos: List[VolumeDTO],
            previous: Dict[str, Chapter] = None
    ):
        volume_mapped_chapters = self.dto_adapter.volumes_from_dto(novel, volume_dtos)

        # add volumes
        self.session.add_all(volume_mapped_chapters.keys())
        self.session.flush()

        # add chapters
        for volume, dtos in volume_mapped_chapters.items():
            chapters = [self.dto_adapter.chapter_from_dto(volume, c) for c in dtos]

            # if we have previous chapters with content, use their content path
            if previous is not None:
                for chapter in chapters:
                    chapter.content = previous.get(chapter.url)

            self.session.add_all(chapters)

        self.session.commit()

    def insert_metadata(self, novel: Novel, metadata_dtos: List[MetaDataDTO]):
        metadata = [self.dto_adapter.metadata_from_dto(novel, c) for c in metadata_dtos]
        self.session.add_all(metadata)
        self.session.commit()

    def update_novel(self, novel: Novel, novel_dto: NovelDTO):
        self.dto_adapter.update_novel_from_dto(novel, novel_dto)
        self.session.commit()

    def set_thumbnail_asset(self, novel: Novel, r_path: Path):
        if novel.thumbnail_path == str(r_path):
            return

        novel.thumbnail_path = str(r_path)
        self.session.commit()

    def update_chapters(self, novel: Novel, volume_dtos: List[VolumeDTO]):
        volumes = self.session.execute(select(Volume).where(Volume.novel_id == novel.id)).scalars().all()
        chapters = self.get_chapters(novel)
        volume_mapped_chapters = self.dto_adapter.volumes_from_dto(novel, volume_dtos)

        indexed_volumes = {v.index: v for v in volumes}
        volumes_to_add = []
        for v in list(volume_mapped_chapters.keys()):
            try:
                existing_volume = indexed_volumes.pop(v.index)

                # update names of volumes that already exist
                if existing_volume.name != v.name:
                    logger.debug(f"Updating volume name (index={v.index}, name='{existing_volume.name}' -> '{v.name}')")
                    self.session.execute(update(Volume).where(Volume.id == existing_volume.id).values(name=v.name))

                # index exists so map the chapters into existing volume
                volume_mapped_chapters[existing_volume] = volume_mapped_chapters.pop(v)
            except KeyError:
                volumes_to_add.append(v)

        # add new volume rows that are new
        logger.debug(f"Adding newly found volumes (count={len(volumes_to_add)})")
        self.session.add_all(volumes_to_add)
        self.session.flush()

        # indexed by url which is unique across table
        indexed_chapters = {c.url: c for c in chapters}
        chapters_to_add = []

        # iterate all new chapters
        for volume, chapter_dtos in volume_mapped_chapters.items():
            for chapter_dto in chapter_dtos:
                chapter = self.dto_adapter.chapter_from_dto(volume, chapter_dto)
                try:
                    # existing chapter equivalent
                    ece = indexed_chapters.pop(chapter.url)

                    # update chapters that need to be updated
                    if ece.volume_id != volume.id \
                            or ece.index != chapter.index:
                        logger.debug(f"Updating chapter parent volume (index={ece.index} -> {chapter.index}, "
                                     f"volume_id={ece.volume_id} -> {volume.id})")
                        self.session.execute(
                            update(Chapter)
                                .where(Chapter.url == chapter.url).values(index=chapter.index, volume_id=volume.id)
                        )
                except KeyError:
                    chapters_to_add.append(chapter)

        # add all new chapters
        logger.debug(f"Adding newly found chapters (count={len(chapters_to_add)}).")
        self.session.add_all(chapters_to_add)

        # delete chapters that dont exist anymore
        logger.debug(f"Deleting chapter rows that dont exist anymore (count={len(indexed_chapters)}).")
        for old in indexed_chapters.values():
            self.session.delete(old)

        # delete volumes that dont exist anymore
        logger.debug(f"Deleting volume rows that dont exist anymore (count={len(indexed_volumes)}).")
        for old in indexed_volumes.values():
            self.session.delete(old)

        self.session.commit()

    def update_metadata(self, novel: Novel, metadata_dtos: List[MetaDataDTO]):
        current_metadata = self.get_metadata(novel)
        indexed_metadata: Dict[Tuple, MetaData] = {(data.name, data.value): data for data in current_metadata}

        metadata_to_add = []
        for dto in metadata_dtos:
            this_metadata = self.dto_adapter.metadata_from_dto(novel, dto)

            try:
                current_metadata = indexed_metadata.pop((this_metadata.name, this_metadata.value))
                if current_metadata.others != this_metadata.others:
                    current_metadata.others = this_metadata
                    logger.debug(f"Updating metadata (id={current_metadata.id}, "
                                 f"others={current_metadata.others} -> {this_metadata.others})")

            except KeyError:
                metadata_to_add.append(this_metadata)

        logger.debug(f"Adding newly found metadata (count={len(metadata_to_add)})")
        self.session.add_all(metadata_to_add)

        logger.debug(f"Deleting metadata rows that dont exist anymore (count={len(indexed_metadata)})")
        for metadata in indexed_metadata.values():
            self.session.delete(metadata)

        self.session.commit()

    def update_content(self, chapter_dto: ChapterDTO):
        stmt = update(Chapter).where(Chapter.url == chapter_dto.url).values(content=chapter_dto.content)
        self.session.execute(stmt)
        self.session.commit()

    def add_url(self, novel: Novel, url: str):
        if url in [novel_url.url for novel_url in self.get_urls(novel)]:
            raise ValueError(f"Url already exists in novel (id={novel.id}, title='{novel.title}', {url=}).")

        novel_url = NovelUrl(url=url, novel_id=novel.id)
        self.session.add(novel_url)
        self.session.commit()

    def remove_url(self, novel: Novel, url: str):
        urls = [novel_url.url for novel_url in self.get_urls(novel)]
        if len(urls) <= 1:
            raise ValueError(f"Novel has only one url (id={novel.id}, title='{novel.title}').\n"
                             f"You must add another before this may be removed.")
        if url not in urls:
            raise ValueError(f"Url not in novel (id={novel.id}, title='{novel.title}', {url=}).")

        stmt = delete(NovelUrl).where((NovelUrl.url == url) & (NovelUrl.novel_id == novel.id))
        self.session.execute(stmt)
        self.session.commit()

    def delete_content(self, novel: Novel):
        volumes = [v.id for v in self.get_volumes(novel)]
        stmt = update(Chapter).where(Chapter.volume_id.in_(volumes)).values(content=None)

        self.session.execute(stmt)
        self.session.commit()

    def delete_novel(self, novel: Novel):
        # using session.delete since we want the sqlalchemy cascade to be run
        self.session.delete(novel)
        self.session.commit()

    def delete_volumes(self, novel: Novel):
        # using session.delete since we want the sqlalchemy cascade to be run
        volumes = self.get_volumes(novel)
        for volume in volumes:
            self.session.delete(volume)

        self.session.commit()

    def delete_metadata(self, novel: Novel):
        stmt = delete(MetaData).where(MetaData.novel_id == novel.id)
        self.session.execute(stmt)
        self.session.commit()














import json
from pathlib import Path

from ...settings import CONFIG_FILE


def from_file() -> dict:
    with CONFIG_FILE.open('r') as f:
        data: dict = json.load(f)

    config = data.get('config', {})

    try:
        config['novel']['dir'] = Path(config['novel']['dir'])
    except KeyError:
        pass

    return config



import unittest
from unittest.mock import patch

from loguru import logger

from novelsave.cli.helpers.source import get_source_gateway
from novelsave.exceptions import SourceNotFoundException


class TestSourceHelper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logger.remove()

    @patch('novelsave.services.source.SourceService')
    def test_get_source_gateway_none(self, source_gateway_provider):
        source_gateway_provider.source_from_url.side_effect = SourceNotFoundException('')

        with self.assertRaises(SystemExit):
            get_source_gateway('https://not.a.provider.site', source_gateway_provider)

    @patch('novelsave.services.source.SourceGateway')
    @patch('novelsave.services.source.SourceService')
    def test_get_source_gateway(self, source_gateway_provider, source_gateway):
        source_gateway_provider.source_from_url.return_value = source_gateway

        result = get_source_gateway('https://not.a.provider.site', source_gateway_provider)
        self.assertEqual(source_gateway_provider.source_from_url.return_value, result)




from functools import lru_cache

import novelsave_sources
import requests
from novelsave_sources import parse_source, parse_metasource, UnknownSourceException

from .meta_source_gateway import MetaSourceGateway
from .source_gateway import SourceGateway
from ...core.services.source import BaseSourceService
from ...exceptions import SourceNotFoundException
from ...utils.adapters import SourceAdapter


class SourceService(BaseSourceService):

    def __init__(self, source_adapter: SourceAdapter):
        self.source_adapter = source_adapter

    @property
    def current_version(self) -> str:
        return novelsave_sources.__version__

    def get_latest_version(self) -> str:
        response = requests.get('https://pypi.org/pypi/novelsave-sources/json')
        if not response.ok:
            raise ConnectionError(f"Response received was not valid: GET {response.url} {response.status_code}")

        data = response.json()
        versions = sorted((v for v in data['releases'].keys() if 'a' not in v), reverse=True)
        return versions[0]

    @lru_cache(1)
    def source_from_url(self, url: str) -> SourceGateway:
        try:
            return SourceGateway(parse_source(url), source_adapter=self.source_adapter)
        except UnknownSourceException:
            raise SourceNotFoundException(url)

    @lru_cache(1)
    def meta_source_from_url(self, url: str) -> MetaSourceGateway:
        try:
            return MetaSourceGateway(parse_metasource(url), source_adapter=self.source_adapter)
        except UnknownSourceException:
            raise SourceNotFoundException(url)



from typing import Tuple, List

import browser_cookie3
from loguru import logger
from novelsave_sources.sources.novel.source import Source
from requests.cookies import RequestsCookieJar

from ...core import dtos
from ...core.services.source import BaseSourceGateway
from ...exceptions import CookieBrowserNotSupportedException
from ...utils.adapters import SourceAdapter


class SourceGateway(BaseSourceGateway):
    def __init__(self, source: Source, source_adapter: SourceAdapter):
        self.source = source
        self.source_adapter = source_adapter

    @property
    def name(self) -> str:
        return getattr(self.source, 'name', type(self.source).__name__)

    @property
    def is_search_capable(self) -> bool:
        return self.source.search_viable

    @property
    def is_login_capable(self) -> bool:
        return self.source.login_viable

    def search(self, keyword: str):
        self.source.search(keyword)

    def login(self, username: str, password: str):
        self.source.login(username, password)

    def novel_by_url(self, url: str) -> dtos.NovelDTO:
        novel = self.source.novel(url)

        return self.source_adapter.novel_to_internal(novel)

    def update_chapter_content(self, chapter: dtos.ChapterDTO) -> dtos.ChapterDTO:
        source_chapter = self.source_adapter.chapter_to_external(chapter)
        self.source.chapter(source_chapter)
        self.source_adapter.chapter_content_to_internal(source_chapter, chapter)

        return chapter

    def use_cookies_from_browser(self, browser: str):
        try:
            cookies = getattr(browser_cookie3, browser)()
        except AttributeError:
            raise CookieBrowserNotSupportedException(browser)

        logger.debug(f"Extracted cookies from browser ({browser=}, count={len(cookies)})")

        cookiejar = self.where_cookies_in_domain(cookies)
        self.source.set_cookies(cookiejar)
        logger.debug(
            f"Filtered and set extracted cookies ({browser=}, source='{type(self).__name__}', count={len(cookiejar)})")

    def where_cookies_in_domain(self, cookies):
        cj = RequestsCookieJar()
        for c in cookies:
            if c.domain in self.source.cookie_domains:
                cj.set(c.name, c.value, domain=c.domain, path=c.path)

        return cj



import unittest

from novelsave.core.dtos import NovelDTO, ChapterDTO, MetaDataDTO
from novelsave.core.entities.novel import Novel, NovelUrl, Chapter, MetaData, Volume
from novelsave.utils.adapters import DTOAdapter


class TestDTOAdapter(unittest.TestCase):

    dto_adapter = DTOAdapter()

    def test_novel_from_dto(self):
        novel_dto = NovelDTO(
            id=None,
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            thumbnail_path=None,
            lang='language',
            url='link',
            last_updated=None,
        )

        expected_novel = Novel(
            title="title",
            author="author",
            synopsis="a nice description",
            thumbnail_url='thumbnail',
            lang='language',
        )

        expected_url = NovelUrl(
            url='link',
        )

        actual_novel, actual_url = self.dto_adapter.novel_from_dto(novel_dto)
        for attrib in {'title', 'author', 'synopsis', 'thumbnail_url', 'lang'}:
            self.assertEqual(getattr(expected_novel, attrib), getattr(actual_novel, attrib))

        for attrib in {'url'}:
            self.assertEqual(getattr(expected_url, attrib), getattr(actual_url, attrib))

    def test_update_novel_from_dto(self):
        novel_dto = NovelDTO(
            id=None,
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            thumbnail_path=None,
            lang='language',
            url='link',
            last_updated=None,
        )

        expected_novel = Novel(
            title="title",
            author="author",
            synopsis="a nice description",
            thumbnail_url='thumbnail',
            lang='language',
        )

        actual_novel = self.dto_adapter.update_novel_from_dto(Novel(), novel_dto)
        for attrib in {'title', 'author', 'synopsis', 'thumbnail_url', 'lang'}:
            self.assertEqual(getattr(expected_novel, attrib), getattr(actual_novel, attrib))

    def test_volume_from_chapter_dtos(self):
        novel = Novel(id=2)

        chapter_dtos = [
            ChapterDTO(index=-1, title='', url='', volume=(1, 'volume')),
            ChapterDTO(index=-1, title='', url='', volume=(1, 'volume')),
            ChapterDTO(index=-1, title='', url='', volume=(1, 'volume')),
            ChapterDTO(index=-1, title='', url='', volume=(2, 'volume 2')),
            ChapterDTO(index=-1, title='', url='', volume=None),
            ChapterDTO(index=1, title='', url='', volume=None),
            ChapterDTO(index=-1, title='', url='', volume=(3,)),
        ]

        expected_volumes = {
            Volume(index=1, name='volume', novel_id=novel.id): chapter_dtos[:3],
            Volume(index=2, name='volume 2', novel_id=novel.id): chapter_dtos[3:4],
            Volume(index=-1, name='_default', novel_id=novel.id): chapter_dtos[4:6],
        }

        actual_volumes = self.dto_adapter.volumes_from_chapter_dtos(novel, chapter_dtos)
        self.assertEqual(len(expected_volumes), len(actual_volumes))
        for (expected_volume, expected_chapter_dtos), (actual_volume, actual_chapter_dtos) in \
                zip(expected_volumes.items(), actual_volumes.items()):
            self.assertListEqual(expected_chapter_dtos, actual_chapter_dtos)
            for attrib in {'id', 'index', 'name', 'novel_id'}:
                self.assertEqual(getattr(expected_volume, attrib), getattr(actual_volume, attrib))

    def test_chapter_from_dto(self):
        volume = Volume(id=1)

        chapter_dto = ChapterDTO(
            index=1,
            title='chapter title',
            url='https://my.chapter.org',
            content='content',
            volume=(1, 'volume'),
        )

        expected_chapter = Chapter(
            id=None,
            index=1,
            title='chapter title',
            url='https://my.chapter.org',
            volume_id=volume.id,
        )

        actual_chapter = self.dto_adapter.chapter_from_dto(volume, chapter_dto)
        for attrib in {'id', 'index', 'title', 'url', 'volume_id'}:
            self.assertEqual(getattr(expected_chapter, attrib), getattr(actual_chapter, attrib))

    def test_chapter_to_dto(self):
        test_chapter = Chapter(
            index=1,
            title="something",
            url="url",
        )

        expected_dto = ChapterDTO(
            index=1,
            title='something',
            url='url',
        )

        actual_dto = self.dto_adapter.chapter_to_dto(test_chapter)
        self.assertEqual(expected_dto, actual_dto)

    def test_metadata_from_dto(self):
        novel = Novel(id=1)

        metadata_dto = MetaDataDTO(
            name='name',
            value='value',
            namespace='ns',
            others={'role': 'this'},
        )

        expected_metadata = MetaData(
            name='name',
            value='value',
            namespace='ns',
            others='{"role": "this"}',
            novel_id=novel.id,
        )

        actual_metadata = self.dto_adapter.metadata_from_dto(novel, metadata_dto)
        for attrib in {'name', 'value', 'namespace', 'others', 'novel_id'}:
            self.assertEqual(getattr(expected_metadata, attrib), getattr(actual_metadata, attrib))


import re
from urllib.parse import urlparse

url_start_pattern = re.compile(r'^https?://')

url_pattern = re.compile(
    r'https?://(www\.)?[-a-zA-Z0-9@:%._+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_+.~#?&/=]*)')


def absolute_url(url: str, location: str):
    """Resolve the url into an absolute url

    :param url: The url to be resolved, it can start with '/', '//', '', or 'https?://'.
    :param location: Location from which the url was retrieved.
    :returns: Absolute url derived using the parameters.
    """
    path = url.lstrip('/')
    if url_start_pattern.match(url):
        return url
    elif url.startswith('/'):
        return f'{location.rstrip("/")}/{path}'

    parse_result = urlparse(location)
    if url.startswith('//'):
        return f'{parse_result.scheme}:{url}'
    else:
        base_url = f'{parse_result.scheme}://{parse_result.netloc}'
        return f'{base_url}/{path}'


def is_url(p_url: str):
    """Checks whether the the provided string is a url

    >>> is_url('http://my.site/')
    True

    >>> is_url('https://www.google.com/')
    True

    >>> is_url('https://site/page')
    False
    """
    return bool(url_pattern.fullmatch(p_url))



import re


def slugify(s: str, replace=''):
    """Remove special characters that will prevent string from being used as file name

    It is important that 'replace' is not a special character.
    Preferably one of '', ' ', '_', or '-'.

    >>> slugify('Hansel&Gretel', replace='-')
    'Hansel-Gretel'

    :param s: String to convert to strip
    :param replace: Character to replace the special characters with
    :returns: String slug that may be safely used as a filename or url part
    """
    return re.sub(r'[\\/:*"\'<>|.%$^&£?]', replace, s)


def format_bytes(size: int):
    """
    scale and label the size of bytes.

    >>> format_bytes(1580)
    '1.54 Kb'

    >>> format_bytes(512)
    '512.00 b'

    barely modified version of https://stackoverflow.com/a/49361727/11728401
    """
    # 2**10 = 1024
    power = 2 ** 10
    n = 0
    power_labels = {0: '', 1: 'K', 2: 'M', 3: 'G', 4: 'T'}
    while size > power:
        size /= power
        n += 1

    return f'{size:.2f} {power_labels[n]}b'



from abc import ABC, abstractmethod
from pathlib import Path


class BaseFileService(ABC):

    @abstractmethod
    def write_str(self, path: Path, data: str):
        """write and replace strings to a file"""

    @abstractmethod
    def write_bytes(self, path: Path, data: bytes):
        """write and replace bytes to a file"""

    @abstractmethod
    def read_str(self, path: Path) -> str:
        """read strings from a file"""

    @abstractmethod
    def read_bytes(self, path: Path) -> bytes:
        """read bytes from a file"""



from ..main import cli
from ... import __version__


@cli.command(name='version')
def _version():
    """Show application version"""

    # Using print here instead of logger because, we want the version
    # as plain possible, also there is no point in printing this
    # to a log file
    print(__version__)



from pathlib import Path

from appdirs import user_config_dir
from tqdm import tqdm

NAME = 'novelsave'
AUTHOR = 'Mensch272'

# base project directory
BASE_DIR = Path(__file__).resolve().parent.parent

# operating system specific configuration file
# config directory is used to place logs, config, cache
CONFIG_DIR = Path(user_config_dir(NAME, AUTHOR))
CONFIG_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_FILE = CONFIG_DIR / 'config.json'

DATA_DIR = CONFIG_DIR / 'data'
DATA_DIR.mkdir(parents=True, exist_ok=True)

DATABASE_FILE = (CONFIG_DIR / 'data.sqlite').resolve()
DATABASE_URL = 'sqlite:///' + str(DATABASE_FILE)

# default novel directory, where packaged files such
# as epub and pdf are stored.
NOVEL_DIR = Path.home() / 'novels'

# the following map defines how files are stored
# by further subdivision into sub-folders
DIVISION_RULES = {
    **{
        key: 'assets'
        for key in {'.jpeg', '.jpg', '.png', '.webp', '.gif'}
    },
}


def console_formatter(record):
    if record['level'].name == 'INFO':
        return '<level>{message}</level>\n'
    else:
        return '<level>{level}: {message}</level>\n'


LOGGER_CONFIG = {
    "handlers": [
        {
            'sink': lambda msg: tqdm.write(msg, end=''),
            'format': console_formatter,
            'level': 'INFO',
            'colorize': True,
            'backtrace': False,
            'diagnose': False,
        },
        {
            'sink': CONFIG_DIR / 'logs' / '{time}.log',
            'level': 'TRACE',
            'retention': '2 days',
            'compression': 'zip',
        },
    ],
}

TQDM_CONFIG = {
    'ncols': 80,
    'bar_format': '{percentage:3.0f}% |{bar}{r_bar}'
}

config = {
    'name': NAME,
    'author': AUTHOR,
    'base_dir': BASE_DIR,
    'config': {
        'dir': CONFIG_DIR,
        'file': CONFIG_FILE,
    },
    'data': {
        'dir': DATA_DIR,
        'division_rules': DIVISION_RULES,
    },
    'novel': {
        'dir': NOVEL_DIR,
    },
    'infrastructure': {
        'database': {
            'url': DATABASE_URL,
        }
    },
    'logger': LOGGER_CONFIG,
    'tqdm': TQDM_CONFIG,
}



from http.cookiejar import CookieJar, Cookie
import unittest
from http.cookies import SimpleCookie
from unittest.mock import patch, MagicMock

from novelsave.exceptions import CookieBrowserNotSupportedException
from novelsave.services.source import SourceGateway


@patch('novelsave.services.source.source_gateway.SourceAdapter')
@patch('novelsave.services.source.source_gateway.Source')
class TestSourceService(unittest.TestCase):

    def test_use_cookies_from_browser_attribute_error(self, source, source_adapter):
        source_gateway = SourceGateway(source, source_adapter)

        with patch(
                'novelsave.services.source.source_gateway.browser_cookie3.chrome',
                side_effect=AttributeError(),
        ):
            with self.assertRaises(CookieBrowserNotSupportedException):
                source_gateway.use_cookies_from_browser('chrome')


from sqlalchemy import Column, Integer, String, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from ..base import Base


class Volume(Base):
    __tablename__ = 'volumes'
    __table_args__ = (
        UniqueConstraint('novel_id', 'index'),
    )

    id = Column(Integer, primary_key=True)
    index = Column(Integer, nullable=False)
    name = Column(String, nullable=False)

    novel_id = Column(Integer, ForeignKey('novels.id'), nullable=False)
    novel = relationship('Novel', back_populates='volumes')

    chapters = relationship('Chapter', back_populates='volume', cascade='all, delete-orphan')



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple

from novelsave.core.entities.novel import Novel


class BasePackager(ABC):

    @abstractmethod
    def keywords(self) -> Tuple[str]:
        """keywords that identify this packager. for example, output format"""

    @abstractmethod
    def package(self, novel: Novel) -> Path:
        """package the a select novel from the database into another format."""

    @abstractmethod
    def destination(self, novel: Novel) -> Path:
        """provide file or directory where the novel has been packaged to, file is preferred."""



from sqlalchemy import Column, Integer, String, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from ..base import Base


class Asset(Base):
    __tablename__ = 'assets'
    __table_args__ = (
        UniqueConstraint('novel_id', 'url'),
    )

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=True)
    url = Column(String)
    path = Column(String)

    type_id = Column(Integer, ForeignKey('asset_types.id'), nullable=False)
    type = relationship('AssetType', back_populates='assets')

    novel_id = Column(Integer, ForeignKey('novels.id'), nullable=False)
    novel = relationship('Novel', back_populates='assets')



from typing import List

from novelsave_sources.sources.metadata.metasource import MetaSource

from novelsave.core.dtos import MetaDataDTO
from novelsave.core.services.source import BaseMetaSourceGateway
from novelsave.utils.adapters import SourceAdapter


class MetaSourceGateway(BaseMetaSourceGateway):

    def __init__(self, meta_source: MetaSource, source_adapter: SourceAdapter):
        self.meta_source = meta_source
        self.source_adapter = source_adapter

    @property
    def name(self):
        return type(self.meta_source).__name__

    def metadata_by_url(self, url: str) -> List[MetaDataDTO]:
        metadata = self.meta_source.retrieve(url)

        return [
            self.source_adapter.metadata_to_internal(m)
            for m in metadata
        ]



import json
from pathlib import Path

from ...core.services.config import BaseConfigService


class ConfigService(BaseConfigService):
    version = 1

    def __init__(self, config_file: Path, default_novel_dir: Path):
        self.data = {'version': self.version, 'config': {}}

        self.config_file = config_file
        self.default_novel_dir = default_novel_dir

        self.config_file.parent.mkdir(parents=True, exist_ok=True)
        self.load()

    def save(self):
        with self.config_file.open('w') as f:
            json.dump(self.data, f, indent=4)

    def load(self):
        try:
            with self.config_file.open('r') as f:
                self.data = json.load(f)
        except FileNotFoundError:
            pass

    @property
    def novel(self) -> dict:
        return self.data['config'].setdefault('novel', {})

    def get_novel_dir(self):
        return self.novel.get('dir', self.default_novel_dir)

    def set_novel_dir(self, path: Path):
        if path.is_file():
            raise NotADirectoryError()

        self.novel['dir'] = str(path)
        self.save()

    def reset_novel_dir(self):
        try:
            del self.novel['dir']
        except KeyError:
            pass

        self.save()



from abc import ABC, abstractmethod
from typing import List, Tuple

from novelsave.core import dtos


class BaseSourceGateway(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        """name of the corresponding source"""

    @property
    @abstractmethod
    def is_search_capable(self) -> bool:
        """identifies whether the source provides search capability"""

    @property
    @abstractmethod
    def is_login_capable(self) -> bool:
        """identifies whether the source provides login capability"""

    @abstractmethod
    def search(self, keyword: str) -> List[dtos.NovelDTO]:
        """search the source library for novels with keyword"""

    @abstractmethod
    def login(self, username: str, password: str):
        """login to the source website, making available services or novels which might otherwise be absent"""

    @abstractmethod
    def novel_by_url(self, url: str) -> dtos.NovelDTO:
        """scrape and parse a novel by its url"""

    @abstractmethod
    def update_chapter_content(self, chapter: dtos.ChapterDTO) -> dtos.ChapterDTO:
        """update a chapter's content by following its url"""

    @abstractmethod
    def use_cookies_from_browser(self, browser: str):
        """take the cookies from the browser and add them to following requests"""



import requests

from novelsave import __version__
from novelsave.core.services import BaseMetaService


class MetaService(BaseMetaService):

    @property
    def current_version(self) -> str:
        return __version__

    def get_latest_version(self) -> str:
        response = requests.get('https://pypi.org/pypi/novelsave/json')
        if not response.ok:
            raise ConnectionError(f"Response received was not valid: GET {response.url} {response.status_code}")

        data = response.json()
        versions = sorted((v for v in data['releases'].keys() if 'a' not in v), reverse=True)
        return versions[0]



from typing import Iterable, Tuple


def unzip_arguments(args: tuple, kwargs: dict, pairs: Iterable[Tuple[int, str]]):
    """
    This is a helper function used extract to arguments from :args and :kwargs
    """
    unzipped = []
    for pair in pairs:
        try:
            unzipped.append(args[pair[0]])
        except IndexError:
            unzipped.append(kwargs[pair[1]])

    return unzipped



import unittest
from pathlib import Path
from unittest.mock import patch

from novelsave.core.entities.novel import Novel
from novelsave.exceptions import SourceNotFoundException
from novelsave.services import PathService


@patch('novelsave.core.services.BaseNovelService')
@patch('novelsave.core.services.source.BaseSourceService')
class TestPathService(unittest.TestCase):
    data_dir = Path('~/data')
    save_dir = Path('~/test_novels')
    division_rules = {'.html': 'web'}

    def test_divide(self, source_service, novel_service):
        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)

        # when there is a specified directory
        r_path = Path('subdivide_test_dir/test_file.html')
        s_path = path_service.divide(r_path)
        self.assertEqual(Path(r_path).parent / 'web' / 'test_file.html', s_path)

        # when there is no specified directory
        r_path = Path('subdivide_test_dir/test_file.json')
        s_path = path_service.divide(r_path)
        self.assertEqual(Path(r_path), s_path)

    def test_get_novel_path_no_source(self, source_service, novel_service):
        source_service.source_from_url.side_effect = SourceNotFoundException('')

        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)
        path = path_service.novel_save_path(Novel(title='novel'))

        self.assertEqual(self.save_dir / 'novel', path)

    @patch('novelsave.core.services.source.BaseSourceGateway')
    def test_get_novel_path_with_source(self, source_service, novel_service, source_gateway):
        source_service.source_from_url.return_value = source_gateway
        source_gateway.name.return_value = 'source'

        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)
        path = path_service.novel_save_path(Novel(title='novel'))

        self.assertEqual(self.save_dir / 'source' / 'novel', path)

    def test_get_thumbnail_path(self, source_service, novel_service):
        novel = Novel(id=1, thumbnail_url='https://my.site/local%20assets/image.jpg')

        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)
        path = path_service.thumbnail_path(novel)

        self.assertEqual(self.data_dir / '1' / 'cover.jpg', path)

    def test_get_thumbnail_path_no_suffix(self, source_service, novel_service):
        novel = Novel(id=1, thumbnail_url='https://my.site/local%20assets/image')

        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)
        path = path_service.thumbnail_path(novel)

        self.assertEqual(self.data_dir / '1' / 'cover', path)

    def test_resolve_data_path(self, source_service, novel_service):
        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)

        test_paths = {
            Path('1/cover.jpg'): self.data_dir / Path('1/cover.jpg'),
            Path('../1/cover.jpg'): self.data_dir / Path('1/cover.jpg'),
            Path('./1/cover.jpg'): self.data_dir / Path('1/cover.jpg'),
        }

        for test_path, expected in test_paths.items():
            path = path_service.resolve_data_path(test_path)
            self.assertEqual(expected, path)

    def test_relative_to_data_dir(self, source_service, novel_service):
        test_path = self.data_dir / Path('file')

        path_service = PathService(self.data_dir, self.save_dir, self.division_rules, novel_service, source_service)
        path = path_service.relative_to_data_dir(test_path)

        self.assertEqual(Path('file'), path)



from dependency_injector import containers, providers
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

from novelsave.services import FileService, NovelService, PathService, AssetService, MetaService
from novelsave.services.packagers import EpubPackager, PackagerProvider
from novelsave.services.config import ConfigService
from novelsave.services.source import SourceService
from novelsave.utils.adapters import SourceAdapter, DTOAdapter


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


class Adapters(containers.DeclarativeContainer):
    source_adapter = providers.Factory(
        SourceAdapter,
    )

    dto_adapter = providers.Factory(
        DTOAdapter,
    )


class Infrastructure(containers.DeclarativeContainer):
    config = providers.Configuration()

    engine = providers.Singleton(create_engine, url=config.database.url, future=True)
    session = providers.Singleton(Session, engine, future=True)


class Services(containers.DeclarativeContainer):
    config = providers.Configuration()

    adapters = providers.DependenciesContainer()
    infrastructure = providers.DependenciesContainer()

    config_service = providers.Singleton(
        ConfigService,
        config_file=config.config.file,
        default_novel_dir=config.novel.dir,
    )

    meta_service = providers.Factory(
        MetaService,
    )

    file_service = providers.Factory(
        FileService,
    )

    novel_service = providers.Factory(
        NovelService,
        session=infrastructure.session,
        dto_adapter=adapters.dto_adapter,
        file_service=file_service,
    )

    source_service = providers.Singleton(
        SourceService,
        source_adapter=adapters.source_adapter,
    )

    path_service = providers.Factory(
        PathService,
        data_dir=config.data.dir,
        novels_dir=config.novel.dir,
        division_rules=config.data.division_rules,
        novel_service=novel_service,
        source_service=source_service,
    )

    asset_service = providers.Factory(
        AssetService,
        session=infrastructure.session,
        path_service=path_service,
    )


class Packagers(containers.DeclarativeContainer):
    novel_config = providers.Configuration()
    services = providers.DependenciesContainer()

    epub_packager = providers.Factory(
        EpubPackager,
        novel_service=services.novel_service,
        file_service=services.file_service,
        path_service=services.path_service,
        asset_service=services.asset_service,
    )

    packager_provider = providers.Factory(
        PackagerProvider,
        epub=epub_packager,
    )


class Application(containers.DeclarativeContainer):
    config = providers.Configuration(strict=True)

    adapters = providers.Container(
        Adapters,
    )

    infrastructure = providers.Container(
        Infrastructure,
        config=config.infrastructure,
    )

    services = providers.Container(
        Services,
        config=config,
        adapters=adapters,
        infrastructure=infrastructure,
    )

    packagers = providers.Container(
        Packagers,
        services=services,
    )



import sys

from dependency_injector.wiring import Provide, inject
from loguru import logger

from novelsave.containers import Application
from novelsave.core.services.source import BaseMetaSourceGateway, BaseSourceGateway, BaseSourceService
from novelsave.exceptions import SourceNotFoundException


@inject
def get_source_gateway(
        url: str,
        source_service: BaseSourceService = Provide[Application.services.source_service],
) -> BaseSourceGateway:
    try:
        source_gateway = source_service.source_from_url(url)
    except SourceNotFoundException:
        logger.error(f"Could not find source corresponding to url ({url=}).")
        sys.exit(1)

    logger.debug(f"Acquired source (name='{source_gateway.name}').")
    return source_gateway


@inject
def get_meta_source_gateway(
        url: str,
        source_service: BaseSourceService = Provide[Application.services.source_service],
) -> BaseMetaSourceGateway:
    try:
        meta_source_gateway = source_service.meta_source_from_url(url)
    except SourceNotFoundException:
        logger.error(f"Could not find metadata source corresponding to url ({url=}).")
        sys.exit(1)

    logger.debug(f"Acquired metadata source (name='{meta_source_gateway.name}').")
    return meta_source_gateway



from sqlalchemy import Column, Integer, String, ForeignKey, TIMESTAMP, func
from sqlalchemy.orm import relationship

from ..base import Base


class Novel(Base):
    __tablename__ = 'novels'

    id = Column(Integer, primary_key=True)
    title = Column(String, nullable=False)
    author = Column(String)
    synopsis = Column(String)
    lang = Column(String)

    thumbnail_url = Column(String)
    thumbnail_path = Column(String)

    last_updated = Column(TIMESTAMP, server_default=func.now(), onupdate=func.current_timestamp())

    urls = relationship('NovelUrl', back_populates='novel', cascade='all, delete-orphan')
    volumes = relationship('Volume', back_populates='novel', cascade='all, delete-orphan')
    novel_metadata = relationship('MetaData', back_populates='novel', cascade='all, delete-orphan')
    assets = relationship('Asset', back_populates='novel', cascade='all, delete-orphan')



import unittest
from unittest.mock import patch

from loguru import logger

from novelsave.cli.controllers import update


@patch('novelsave.cli.controllers._update.helpers.download_assets')
@patch('novelsave.cli.controllers._update.helpers.download_chapters')
@patch('novelsave.cli.controllers._update.helpers.download_thumbnail')
@patch('novelsave.cli.controllers._update.helpers.update_novel')
@patch('novelsave.cli.controllers._update.helpers.create_novel', return_value=None)
@patch('novelsave.cli.controllers._update.helpers.get_novel', return_value=None)
class TestUpdateController(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logger.remove()

    def test_update_not_exists(self, get_novel, create_novel, update_novel, download_thumbnail, download_pending, download_assets):
        get_novel.side_effect = ValueError()
        create_novel.return_value = None

        with self.assertRaises(SystemExit):
            update('not int', 0, None, None)
        get_novel.assert_called_with('not int')
        create_novel.assert_not_called()

        with self.assertRaises(SystemExit):
            update('1', 0, None, None)
        get_novel.assert_called_with('1')
        create_novel.assert_not_called()

        update('https://my.site.org', None, 0, None)
        get_novel.assert_called_with('https://my.site.org')
        create_novel.assert_called_with('https://my.site.org', None)

    def test_update_exists(self, get_novel, create_novel, update_novel, download_thumbnail, download_pending, download_assets):
        get_novel.return_value = 'novel'

        url = 'https://novel.site'
        update(url, None, 1, None)

        get_novel.assert_called_with(url)
        update_novel.assert_called_with('novel', None)
        download_pending.assert_called_with('novel', 1, None)

    def test_update_exists_skip_chapters(self, get_novel, create_novel, update_novel, download_thumbnail,
                                         download_pending, download_assets):
        get_novel.return_value = 'novel'

        url = 'https://novel.site'
        update(url, None, limit=0, threads=None)

        get_novel.assert_called_with(url)
        update_novel.assert_called_with('novel', None)
        download_pending.assert_not_called()




import sys

from dependency_injector.wiring import inject, Provide
from loguru import logger

from novelsave.cli.helpers import get_novel
from novelsave.containers import Application
from novelsave.core.services import BasePathService
from novelsave.core.services.packagers import BasePackagerProvider


@inject
def package(
        id_or_url: str,
        packager_provider: BasePackagerProvider = Provide[Application.packagers.packager_provider],
        path_service: BasePathService = Provide[Application.services.path_service],
):
    """
    package the selected novel into the formats of choosing

    Note: currently supports epub format only

    :return: None
    """
    try:
        novel = get_novel(id_or_url)
    except ValueError:
        sys.exit(1)

    for packager in packager_provider.packagers():
        path = packager.package(novel)
        logger.info(f"Packaged (type='{packager.keywords()[0]}', path='{path_service.relative_to_novel_dir(path)}')")



import unittest

import tempfile
import shutil
from pathlib import Path

from novelsave.services.file_service import FileService


class TestFileService(unittest.TestCase):

    temp_dir = Path(tempfile.gettempdir())
    data_division = {'.html': 'web'}
    file_service = FileService()

    def test_write_str(self):
        path = Path('ns_test_dir/atc_file.html')
        path.parent.mkdir(parents=True, exist_ok=True)

        data = '<h1>testing data</h1>'
        self.file_service.write_str(path, data)

        # test file was created
        self.assertTrue(path.exists(), 'path does not exist')
        self.assertTrue(path.is_file(), 'path is not a file')

        # test content was written to file
        with path.open('r') as f:
            self.assertEqual(data, f.read())

        # cleanup: remove created dir
        shutil.rmtree(path.parent)

    def test_write_bytes(self):
        path = Path('ns_test_dir/atc_file.html')
        path.parent.mkdir(parents=True, exist_ok=True)

        data = b'<h1>testing data</h1>'
        self.file_service.write_bytes(path, data)

        # test file was created
        self.assertTrue(path.exists(), 'path does not exist')
        self.assertTrue(path.is_file(), 'path is not a file')

        # test content was written to file
        with path.open('rb') as f:
            self.assertEqual(data, f.read())

        # cleanup: remove created dir
        shutil.rmtree(path.parent)

    def test_read_str(self):
        path = Path('ns_test_dir/atc_file.html')
        path.parent.mkdir(parents=True, exist_ok=True)

        data = '<h1>testing data read</h1>'

        # create file and write data
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open('w') as f:
            f.write(data)

        # read and test written data
        read_data = self.file_service.read_str(path)
        self.assertEqual(data, read_data)

        # cleanup: remove created dir
        shutil.rmtree(path.parent)

    def test_read_str_no_file(self):
        # non existent file
        with self.assertRaises(FileNotFoundError):
            self.file_service.read_str(Path('ns_ne_dir/non.html'))

        # throw when its not a file
        dir = Path('ns_ne_n_file')
        path = self.temp_dir / dir
        path.mkdir(parents=True, exist_ok=True)
        with self.assertRaises(FileNotFoundError):
            self.file_service.read_str(dir)

        # cleanup: remove created dir
        shutil.rmtree(path)

    def test_read_bytes(self):
        path = Path('ns_test_dir/atc_file.html')
        data = b'<h1>testing data read</h1>'

        # create file and write data
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open('wb') as f:
            f.write(data)

        # read and test written data
        read_data = self.file_service.read_bytes(path)
        self.assertEqual(data, read_data)

        # cleanup: remove created dir
        shutil.rmtree(path.parent)

    def test_read_bytes_no_file(self):
        # non existent file
        with self.assertRaises(FileNotFoundError):
            self.file_service.read_bytes(Path('ns_ne_dir/non.html'))

        # throw when its not a file
        dir = Path('ns_ne_n_file')
        path = self.temp_dir / dir
        path.mkdir(parents=True, exist_ok=True)
        with self.assertRaises(FileNotFoundError):
            self.file_service.read_bytes(dir)

        # cleanup: remove created dir
        shutil.rmtree(path)



from abc import ABC, abstractmethod
from typing import Dict, List

from novelsave.core.dtos import ChapterDTO
from novelsave.core.entities.novel import Novel, Asset


class BaseAssetService(ABC):

    @abstractmethod
    def downloaded_assets(self, novel: Novel) -> List[Asset]:
        """all assets belonging to novel that have been downloaded"""

    @abstractmethod
    def pending_assets(self, novel: Novel) -> List[Asset]:
        """assets that have not been downloaded"""

    @abstractmethod
    def update_asset_path(self, asset: Asset):
        """update the file path of the asset"""

    @abstractmethod
    def delete_assets_of_novel(self, novel: Novel):
        """delete all assets that are associated with from database"""

    @abstractmethod
    def collect_assets(self, novel: Novel, chapter: ChapterDTO) -> str:
        """collect and modify the provided the html for asset injection"""

    @abstractmethod
    def mapping_dict(self, path_mapping: Dict[int, str]):
        """create a mapping dict that can be used by inject assets"""

    @abstractmethod
    def inject_assets(self, html: str, mapping_dict: Dict[int, str]):
        """replace the placeholder links with new link to assets"""



import os
import shutil
import sys
from concurrent import futures
from functools import lru_cache
from typing import Optional

import requests
from dependency_injector.wiring import inject, Provide
from loguru import logger
from tqdm import tqdm

from novelsave.cli.helpers.source import get_source_gateway
from novelsave.containers import Application
from novelsave.core.dtos import ChapterDTO
from novelsave.core.entities.novel import Novel
from novelsave.core.services import BasePathService, BaseNovelService, BaseAssetService, BaseFileService
from novelsave.core.services.source import BaseSourceGateway
from novelsave.exceptions import ContentUpdateFailedException, NSError
from novelsave.exceptions import CookieBrowserNotSupportedException
from novelsave.settings import TQDM_CONFIG
from novelsave.utils.adapters import DTOAdapter
from novelsave.utils.helpers import url_helper, string_helper


def set_cookies(source_gateway: BaseSourceGateway, browser: Optional[str]):
    """sets the cookies in source gateway, process skipped if browser is none"""
    if browser is None:
        return

    try:
        source_gateway.use_cookies_from_browser(browser)
    except CookieBrowserNotSupportedException:
        logger.error(f"Extracting cookies from browser not supported. ({browser=})")
        sys.exit(1)

    logger.info(f"Applied cookies from browser ({browser=}).")


def retrieve_novel_info(source_gateway: BaseSourceGateway, url: str, browser: str):
    set_cookies(source_gateway, browser)

    logger.info(f"Retrieving novel information ({url=})...")
    try:
        output = source_gateway.novel_by_url(url)
    except requests.ConnectionError:
        raise NSError(f"Connection terminated unexpectedly; Make sure you are connected to the internet.")

    return output


@inject
def create_novel(
        url: str,
        browser: Optional[str],
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        path_service: BasePathService = Provide[Application.services.path_service],
) -> Novel:
    """
    retrieve information about the novel from webpage and insert novel into database.
    this includes chapter list and metadata.
    """
    source_gateway = get_source_gateway(url)
    novel_dto = retrieve_novel_info(source_gateway, url, browser)

    novel = novel_service.insert_novel(novel_dto)
    novel_service.insert_chapters(novel, novel_dto.volumes)
    novel_service.insert_metadata(novel, novel_dto.metadata)

    chapters = [c for v in novel_dto.volumes for c in v.chapters]
    logger.info(f"Added new novel (id={novel.id}, title='{novel.title}', chapters={len(chapters)}').")

    data_dir = path_service.novel_data_path(novel)
    if data_dir.exists():
        shutil.rmtree(data_dir)
        logger.debug(f"Cleaned existing data in novel data dir (path='{path_service.relative_to_data_dir(data_dir)}')")

    return novel


@inject
def update_novel(
        novel: Novel,
        browser: Optional[str],
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
):
    url = novel_service.get_primary_url(novel)
    logger.debug(f"Using primary url ({url=})")

    source_gateway = get_source_gateway(url)
    novel_dto = retrieve_novel_info(source_gateway, url, browser)

    novel_service.update_novel(novel, novel_dto)
    novel_service.update_chapters(novel, novel_dto.volumes)
    novel_service.update_metadata(novel, novel_dto.metadata)

    chapters = [c for v in novel_dto.volumes for c in v.chapters]
    logger.info(f"Updated novel (id={novel.id}, title='{novel.title}', chapters={len(chapters)})")
    return novel


@inject
def download_thumbnail(
        novel: Novel,
        force: bool = False,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        file_service: BaseFileService = Provide[Application.services.file_service],
        path_service: BasePathService = Provide[Application.services.path_service],
):
    thumbnail_path = path_service.thumbnail_path(novel)
    novel_service.set_thumbnail_asset(novel, path_service.relative_to_data_dir(thumbnail_path))

    if not force and thumbnail_path.exists() and thumbnail_path.is_file():
        logger.info(f"Skipped thumbnail download (title='{novel.title}', reason='File already exists')")
        return

    logger.debug(f"Attempting thumbnail download (title='{novel.title}', thumbnail='{novel.thumbnail_path}').")
    try:
        response = requests.get(novel.thumbnail_url)
    except requests.ConnectionError as e:
        raise NSError(f"Connection terminated unexpectedly; Make sure you are connected to the internet.")

    if not response.ok:
        logger.error(f"Error during thumbnail download (title='{novel.title=}', {response.status_code=}).")
        return

    thumbnail_path.parent.mkdir(parents=True, exist_ok=True)
    file_service.write_bytes(thumbnail_path, response.content)

    size = string_helper.format_bytes(len(response.content))
    logger.info(f"Downloaded thumbnail image (title='{novel.title}', thumbnail='{novel.thumbnail_path}', {size=}).")


@inject
def download_chapters(
        novel: Novel,
        limit: Optional[int],
        threads: Optional[int],
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
        asset_service: BaseAssetService = Provide[Application.services.asset_service],
        dto_adapter: DTOAdapter = Provide[Application.adapters.dto_adapter],
):
    chapters = novel_service.get_pending_chapters(novel, limit)
    if not chapters:
        logger.info(f"Skipped chapter download (title='{novel.title}', reason='No pending chapters').")
        return

    url = novel_service.get_primary_url(novel)
    logger.debug(f"Using primary novel url ({url=}).")

    source_gateway = get_source_gateway(url)
    thread_count = min(threads, os.cpu_count())

    def download(dto: ChapterDTO):
        try:
            return source_gateway.update_chapter_content(dto)
        except Exception as e:
            raise ContentUpdateFailedException(dto, e)

    logger.info(f"Downloading pending chapters (count={len(chapters)}, threads={thread_count})...")
    successes = 0
    with tqdm(total=len(chapters), **TQDM_CONFIG) as pbar:
        executor = futures.ThreadPoolExecutor(max_workers=thread_count)
        download_futures = [executor.submit(download, dto_adapter.chapter_to_dto(c)) for c in chapters]

        for future in futures.as_completed(download_futures):
            if future.done():
                chapter_dto = future.result()
                chapter_dto.content = asset_service.collect_assets(novel, chapter_dto)
                novel_service.update_content(chapter_dto)

                logger.debug(f"Chapter content downloaded (index={chapter_dto.index}, title='{chapter_dto.title}').")
                successes += 1
            else:
                exc = future.exception()
                logger.error(f"An error occurred during content download "
                             f"(title='{exc.chapter.title}', error={type(exc.exception)}).")
                logger.debug("An error occurred during content download {}", vars(exc))

            pbar.update(1)

        executor.shutdown()

    logger.info(f"Chapters download complete (successes={successes}, errors={len(chapters) - successes}).")

@inject
def download_assets(
        novel: Novel,
        asset_service: BaseAssetService = Provide[Application.services.asset_service],
        file_service: BaseFileService = Provide[Application.services.file_service],
        path_service: BasePathService = Provide[Application.services.path_service],
):
    pending = asset_service.pending_assets(novel)
    if not pending:
        logger.info(f"Skipped assets download (title='{novel.title}', reason='No pending assets').")
        return

    logger.info(f"Downloading pending assets (count={len(pending)}).")
    for asset in tqdm(pending, **TQDM_CONFIG):
        response = requests.get(asset.url)
        if not response.ok:
            logger.error(f"Error during asset download (id={asset.id}, url={asset.url}).")
            continue

        file = path_service.asset_path(novel, asset)
        file.parent.mkdir(parents=True, exist_ok=True)

        file_service.write_bytes(file, response.content)

        asset.path = str(path_service.relative_to_data_dir(file))
        asset_service.update_asset_path(asset)

        logger.debug(f"Asset downloaded and saved (id={asset.id}, url={asset.url}, path={asset.path}).")

    logger.info(f"Assets download complete.")


@lru_cache(maxsize=1)
@inject
def get_novel(
        id_or_url: str,
        silent: bool = False,
        novel_service: BaseNovelService = Provide[Application.services.novel_service],
) -> Novel:
    """retrieve novel is it exists in the database otherwise return none

    :raises ValueError: if novel does not exist
    """

    is_url = url_helper.is_url(id_or_url)
    if is_url:
        novel = novel_service.get_novel_by_url(id_or_url.rstrip('/'))
    else:
        try:
            novel = novel_service.get_novel_by_id(int(id_or_url))
        except ValueError:
            logger.error(f"Value provided is neither a url or an id (value={({id_or_url})}).")
            sys.exit(1)

    if not novel:
        quote = "'" if is_url else ''
        logger.error(f"Novel not found ({'url' if is_url else 'id'}={quote}{id_or_url}{quote}).")
        raise ValueError("Novel was not found.")

    if not silent:
        logger.info(f"Acquired novel from database (id={novel.id}, title='{novel.title}').")

    return novel


@inject
def get_or_create_novel(
        id_or_url: str,
) -> Novel:
    """retrieve specified novel from database or web-crawl and create the novel"""
    novel = get_novel(id_or_url)

    if novel is None:
        if not id_or_url.startswith("http"):
            sys.exit(1)

        novel = create_novel(id_or_url)

    return novel



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List

from .metadata_dto import MetaDataDTO
from .volume_dto import VolumeDTO


@dataclass
class NovelDTO:
    id: Optional[int]
    title: str
    url: str
    author: str = None
    synopsis: str = None

    thumbnail_url: str = None
    thumbnail_path: str = None

    lang: str = 'en'
    last_updated: datetime = None

    volumes: List[VolumeDTO] = field(default_factory=lambda: [])
    metadata: List[MetaDataDTO] = field(default_factory=lambda: [])



from pathlib import Path

import click
from dependency_injector.wiring import inject, Provide
from loguru import logger

from ..main import cli
from ...containers import Application
from ...core.services.config import BaseConfigService


@cli.group(name='config')
def _config():
    """Manage customizable configurations"""


@_config.group(name='novel_dir')
def _novel_dir():
    """Manage configuration novel save location"""


@_novel_dir.command(name='set')
@click.argument('path')
@inject
def _set_novel_dir(path, config_service: BaseConfigService = Provide[Application.services.config_service]):
    """Set novel save location"""
    config_service.set_novel_dir(Path(path))
    logger.info(f"Novel save directory has been updated ({path=})")


@_novel_dir.command(name='reset')
@inject
def _reset_novel_dir(config_service: BaseConfigService = Provide[Application.services.config_service]):
    """Reset novel save location to default"""
    config_service.reset_novel_dir()
    logger.info(f"Novel save directory has been reset (path={config_service.get_novel_dir()})")



import json
from pathlib import Path
from typing import Tuple, List, Dict

from novelsave.core.dtos import NovelDTO, ChapterDTO, MetaDataDTO, VolumeDTO
from novelsave.core.entities.novel import Novel, NovelUrl, Chapter, Volume, MetaData


class DTOAdapter:

    def novel_from_dto(self, novel_dto: NovelDTO) -> Tuple[Novel, NovelUrl]:
        novel = Novel(
            title=novel_dto.title,
            author=novel_dto.author,
            synopsis=novel_dto.synopsis,
            thumbnail_url=novel_dto.thumbnail_url,
            lang=novel_dto.lang,
        )

        url = NovelUrl(
            # ensure trailing forward slash, it is preferred that a uri that isn't
            # a file end in a slash to signify a path
            url=novel_dto.url.rstrip('/') + ('/' if Path(novel_dto.url).suffix else ''),
        )

        return novel, url

    def update_novel_from_dto(self, novel: Novel, novel_dto: NovelDTO) -> Novel:
        novel.title = novel_dto.title
        novel.author = novel_dto.author
        novel.synopsis = novel_dto.synopsis
        novel.thumbnail_url = novel_dto.thumbnail_url
        novel.lang = novel_dto.lang

        return novel

    def volumes_from_dto(self, novel: Novel, volume_dtos: List[VolumeDTO]) -> Dict[Volume, List[ChapterDTO]]:
        return {
            Volume(
                id=dto.id,
                index=dto.index,
                name=dto.name,
                novel_id=novel.id,
            ): dto.chapters
            for dto in volume_dtos
        }

    def chapter_from_dto(self, volume: Volume, chapter_dto: ChapterDTO) -> Chapter:
        return Chapter(
            index=chapter_dto.index,
            title=chapter_dto.title,
            url=chapter_dto.url,
            volume_id=volume.id,
        )

    def chapter_to_dto(self, chapter: Chapter):
        return ChapterDTO(
            index=chapter.index,
            title=chapter.title,
            url=chapter.url,
        )

    def metadata_from_dto(self, novel: Novel, metadata_dto: MetaDataDTO) -> MetaData:
        return MetaData(
            name=metadata_dto.name,
            value=metadata_dto.value,
            namespace=metadata_dto.namespace,
            others=json.dumps(metadata_dto.others),
            novel_id=novel.id,
        )



import sys
from typing import Optional

from loguru import logger

from novelsave.utils.helpers import url_helper
from .. import helpers


def update(
        id_or_url: str,
        browser: Optional[str],
        limit: Optional[int],
        threads: Optional[int],
):
    """
    update the novel metadata and downloads any new chapters if not specified otherwise

    if url is provided and novel does not exist, a new novel entry will be created.
    however if id is provided the process will be terminated

    :param id_or_url: id or url of the novel
    :param browser: extract cookies from this browser
    :param limit: no. of chapters to update
    :param threads: no. of threads to use when downloading chapters
    :return: None
    """
    try:
        novel = helpers.get_novel(id_or_url)
    except ValueError:
        is_url = url_helper.is_url(id_or_url)
        if not is_url:
            sys.exit(1)

        logger.info(f"Attempting to create new novel entry using url.")
        novel = helpers.create_novel(id_or_url, browser)
    else:
        helpers.update_novel(novel, browser)

    helpers.download_thumbnail(novel)

    if limit is None or limit > 0:
        helpers.download_chapters(novel, limit, threads)
    else:
        logger.info(f"Skipped chapter download (title='{novel.title}', reason='No download specified').")

    helpers.download_assets(novel)



from abc import ABC, abstractmethod
from typing import List, Set

from novelsave.core.services.packagers import BasePackager


class BasePackagerProvider(ABC):

    @abstractmethod
    def keywords(self):
        """all keywords of the supported packagers"""

    @abstractmethod
    def packagers(self) -> Set[BasePackager]:
        """return all the packagers"""

    @abstractmethod
    def filter_packagers(self, keywords: List[str]) -> Set[BasePackager]:
        """return all packagers with the one of the specified keywords"""




from abc import ABC, abstractmethod

from . import BaseSourceGateway, BaseMetaSourceGateway


class BaseSourceService(ABC):

    @property
    @abstractmethod
    def current_version(self) -> str:
        """installed current_version of sources"""

    @abstractmethod
    def get_latest_version(self) -> str:
        """Latest stable version available for download"""

    @abstractmethod
    def source_from_url(self, url: str) -> BaseSourceGateway:
        """find source gateway pointing towards source that handles the url specified

        :raises NovelSourceNotFoundException: if no source if available for the url
        """

    @abstractmethod
    def meta_source_from_url(self, url: str) -> BaseMetaSourceGateway:
        """find metadata source gateway pointing towards source that's handles the url specified

        :raises MetaDataSourceNotFoundException: if no source if available for the url
        """



import atexit

import requests
from dependency_injector.wiring import inject, Provide
from loguru import logger

from novelsave.containers import Application
from novelsave.core.services import BaseMetaService
from novelsave.core.services.source import BaseSourceService


@atexit.register
@inject
def update_check_event(
        source_service: BaseSourceService = Provide[Application.services.source_service],
        meta_service: BaseMetaService = Provide[Application.services.meta_service],
):
    logger.debug("Checking for new package versions...")

    if type(source_service) == Provide:
        logger.debug("Service injection failed. Exiting process silently.")
        return

    available_updates = []
    errors = []

    try:   # novelsave-sources
        latest_sources_version = source_service.get_latest_version()
        if latest_sources_version > source_service.current_version:
            available_updates.append(('novelsave-sources', latest_sources_version))
    except requests.ConnectionError as e:
        errors.append(e)
        logger.debug(f"Connection terminated unexpectedly while checking for update (package='novelsave-sources').")

    try:  # novelsave
        latest_version = meta_service.get_latest_version()
        if latest_version > meta_service.current_version:
            available_updates.append(('novelsave', latest_version))
    except requests.ConnectionError as e:
        errors.append(e)
        logger.debug(f"Connection terminated unexpectedly while checking for update (package='novelsave').")

    if not available_updates:
        logger.debug(f'No upgrades to packages detected (errors={len(errors)}).')
        return

    logger.warning(
        f"Following packages have new versions available: {', '.join(f'{n}=={v}' for n, v in available_updates)}.\n"
        f"Run the following command to upgrade:\n"
        f"   python -m pip install --upgrade {' '.join(n for n, v in available_updates)}")



from typing import List, Set, Iterable, Tuple

from novelsave.core.services.packagers import BasePackager, BasePackagerProvider


class PackagerProvider(BasePackagerProvider):

    def __init__(self, epub: BasePackager):
        self._packagers = (epub, )

    def keywords(self):
        return [keyword for p in self._packagers for keyword in p.keywords()]

    def packagers(self) -> Tuple[BasePackager]:
        return self._packagers

    def filter_packagers(self, keywords: List[str]) -> Set[BasePackager]:
        return {
            compiler
            for compiler in self._packagers
            if any(k in keywords for k in compiler.keywords())
        }



from .novel import Novel
from .novel_url import NovelUrl
from .volume import Volume
from .chapter import Chapter
from .asset import Asset
from .asset_type import AssetType
from .metadata import MetaData



from ._novel import \
    _package, _process, _update, _novel, _info, _clean_novel, _delete_novel, _url, _add_url, _remove_url
# from ._manage import _manage
from ._config import _set_novel_dir
from ._version import _version



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, List, Dict

from novelsave.core.dtos import NovelDTO, ChapterDTO, MetaDataDTO, VolumeDTO
from novelsave.core.entities.novel import Novel, Chapter, Volume, MetaData, NovelUrl


class BaseNovelService(ABC):

    @abstractmethod
    def get_all_novels(self) -> List[Novel]:
        """retrieve all the novels in the database"""

    @abstractmethod
    def get_novel_by_id(self, id_: int) -> Optional[Novel]:
        """retrieve novel by its id, assuming it exists, otherwise return none"""

    @abstractmethod
    def get_novel_by_url(self, url: str) -> Optional[Novel]:
        """retrieve novel by url, if novel doesn't exist return none"""

    @abstractmethod
    def get_primary_url(self, novel: Novel) -> str:
        """retrieve the primary url corresponding to the novel"""

    @abstractmethod
    def get_urls(self, novel) -> List[NovelUrl]:
        """retrieve all the urls of novel"""

    @abstractmethod
    def get_chapters(self, novel: Novel) -> List[Chapter]:
        """retrieve all chapters of the novel"""

    @abstractmethod
    def get_pending_chapters(self, novel: Novel, limit: int) -> List[Chapter]:
        """retrieve all pending chapters of the novel. a chapter is assumed to be pending if it has no content."""

    @abstractmethod
    def get_volumes(self, novel: Novel) -> List[Volume]:
        """retrieve all the volumes of the novel"""

    @abstractmethod
    def get_volumes_with_chapters(self, novel: Novel) -> Dict[Volume, List[Chapter]]:
        """retrieve all volumes along with their corresponding chapters that have content"""

    @abstractmethod
    def get_metadata(self, novel: Novel) -> List[MetaData]:
        """retrieve all metadata of novel"""

    @abstractmethod
    def insert_novel(self, novel_dto: NovelDTO) -> Novel:
        """insert a new novel into the database"""

    @abstractmethod
    def insert_chapters(self, novel: Novel, volume_dtos: List[VolumeDTO]):
        """insert new chapters into the database as well as their volumes"""

    @abstractmethod
    def insert_metadata(self, novel: Novel, metadata_dtos: List[MetaDataDTO]):
        """insert new metadata into the database"""

    @abstractmethod
    def set_thumbnail_asset(self, novel: Novel, r_path: Path):
        """update the cover asset path of the novel"""

    @abstractmethod
    def update_novel(self, novel: Novel, novel_dto: NovelDTO):
        """update existing novel with newer values"""

    @abstractmethod
    def update_chapters(self, novel: Novel, volume_dtos: List[VolumeDTO]):
        """update and replace existing novel chapters and their corresponding volumes"""

    @abstractmethod
    def update_metadata(self, novel: Novel, metadata_dtos: List[MetaDataDTO]):
        """update and replace existing metadata"""

    @abstractmethod
    def update_content(self, chapter_dto: ChapterDTO):
        """update an existing chapter's content"""

    @abstractmethod
    def add_url(self, novel: Novel, url: str):
        """add the specified url to the database linked to novel"""

    @abstractmethod
    def remove_url(self, novel: Novel, url: str):
        """removes the specified url from the database"""

    @abstractmethod
    def delete_content(self, novel: Novel):
        """deletes all the chapters contents"""

    @abstractmethod
    def delete_novel(self, novel: Novel):
        """delete novel including chapters and assets"""

    @abstractmethod
    def delete_volumes(self, novel: Novel):
        """delete volumes associated with novel as well as the chapters linked to them"""

    @abstractmethod
    def delete_metadata(self, novel: Novel):
        """delete all the metadata associated with novel"""



import unittest
from unittest.mock import Mock

from novelsave.core.entities.novel import MetaData
from novelsave.services.packagers import EpubPackager


class TestEpubCompiler(unittest.TestCase):

    def test_metadata_display_value(self):
        compiler = EpubPackager(Mock(), Mock(), Mock(), Mock())

        data = MetaData(name='name', value='value', namespace='DC', others='{"key":"value"}')
        value = compiler.metadata_display_value(data)
        self.assertEqual('value(key=value)', value)

        data = MetaData(name='name', value='value', namespace='DC', others='{}')
        value = compiler.metadata_display_value(data)
        self.assertEqual('value', value)



from novelsave_sources import models as sm
from novelsave.core import dtos


class SourceAdapter(object):
    """adapter responsible for converting models from view_models to source and vice versa"""

    def novel_to_internal(self, novel: sm.Novel) -> dtos.NovelDTO:
        """convert novel from internal to source"""

        arguments = {
            key: value
            for key, value in vars(novel).items()
            if key in {'title', 'url', 'author', 'thumbnail_url', 'lang'}
        }

        return dtos.NovelDTO(
            id=None,
            synopsis='\n'.join(novel.synopsis),
            volumes=[self.volume_to_internal(c) for c in novel.volumes],
            metadata=[self.metadata_to_internal(m) for m in novel.metadata],
            **arguments
        )

    def novel_to_external(self, novel: dtos.NovelDTO) -> sm.Novel:
        """convert novel from source to internal"""

        arguments = {
            key: value
            for key, value in vars(novel).items()
            if key in {'title', 'url', 'author', 'synopsis', 'thumbnail_url', 'lang'}
        }

        return sm.Novel(**arguments)

    def volume_to_internal(self, volume: sm.Volume) -> dtos.VolumeDTO:
        """convert volume from source to internal"""

        return dtos.VolumeDTO(
            id=None,
            index=volume.index,
            name=volume.name,
            chapters=[self.chapter_to_internal(c) for c in volume.chapters]
        )

    def chapter_to_internal(self, chapter: sm.Chapter) -> dtos.ChapterDTO:
        """convert chapter from source to internal"""

        return dtos.ChapterDTO(
            index=chapter.index,
            title=chapter.title,
            url=chapter.url,
            content=chapter.paragraphs,
        )

    def chapter_to_external(self, chapter: dtos.ChapterDTO) -> sm.Chapter:
        """convert chapter from internal to source"""

        return sm.Chapter(
            index=chapter.index,
            title=chapter.title,
            url=chapter.url,
            paragraphs=chapter.content,
        )

    def metadata_to_internal(self, metadata: sm.Metadata) -> dtos.MetaDataDTO:
        """convert metadata from source to internal"""

        return dtos.MetaDataDTO(
            name=metadata.name,
            value=metadata.value,
            namespace=metadata.namespace,
            others=metadata.others,
        )

    def metadata_to_external(self, metadata: dtos.MetaDataDTO) -> sm.Metadata:
        """convert metadata from internal to source"""

        return sm.Metadata(
            name=metadata.name,
            value=metadata.value,
            namespace=metadata.namespace,
            others=metadata.others,
        )

    def chapter_content_to_internal(self, source_chapter: sm.Chapter, internal_chapter: dtos.ChapterDTO):
        """map content from source chapter to internal chapter"""
        internal_chapter.content = source_chapter.paragraphs



import unittest

import novelsave.utils.helpers.url_helper
from novelsave.utils.helpers import string_helper


class TestStringHelper(unittest.TestCase):

    def test_is_url_true(self):
        tests = [
            'https://github.com/mHaisham/novelsave/',
            'https://www.webnovel.com/book/lord-of-mysteries_11022733006234505',
            'https://www.scribblehub.com/series/10700/tree-of-aeons-an-isekai-story/'
            'https://www.royalroad.com/fiction/21410/super-minion',
        ]

        for url in tests:
            self.assertTrue(novelsave.utils.helpers.url_helper.is_url(url), msg=f"{url=} is not a url")

    def test_is_url_false(self):
        tests = [
            'https:/github.com',
            'this',
            '1',
            '1.001',
            'https://1',
        ]

        for url in tests:
            self.assertFalse(novelsave.utils.helpers.url_helper.is_url(url), msg=f"{url=} is a url")



from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from ..base import Base


class AssetType(Base):
    __tablename__ = 'asset_types'

    id = Column(Integer, primary_key=True)
    name = Column(String(16), nullable=False)

    assets = relationship('Asset', back_populates='type')





from dataclasses import dataclass, field


@dataclass
class MetaDataDTO:
    DEFAULT_NAMESPACE = 'DC'
    SOURCE_INTERNAL = 'int'
    SOURCE_EXTERNAL = 'ext'

    name: str
    value: str
    others: dict = field(default_factory=lambda: {})
    namespace: str = DEFAULT_NAMESPACE
    src: str = SOURCE_INTERNAL



import unittest
from unittest.mock import Mock

from novelsave_sources.sources.novel import sources

from novelsave.exceptions import SourceNotFoundException
from novelsave.services.source import SourceService


class TestSourceServiceProvider(unittest.TestCase):
    source_service = SourceService(Mock())

    def test_source_from_url(self):
        source_service = self.source_service.source_from_url(sources[0].base_urls[0])
        self.assertIsNotNone(source_service)

    def test_source_from_url_unavailable(self):
        with self.assertRaises(SourceNotFoundException):
            self.source_service.source_from_url('https://test.site')



import unittest
from unittest.mock import MagicMock, patch

from loguru import logger

from novelsave.cli.helpers import novel as novel_helper
from novelsave.core.entities.novel import Novel
from novelsave.exceptions import CookieBrowserNotSupportedException


class TestNovelHelper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logger.remove()

    def test_set_cookies(self):
        with patch('novelsave.cli.helpers.novel.BaseSourceGateway') as source_gateway:
            source_gateway.use_cookies_from_browser.side_effect = CookieBrowserNotSupportedException('')
            novel_helper.set_cookies(source_gateway, None)
            source_gateway.use_cookies_from_browser.assert_not_called()

        with patch('novelsave.cli.helpers.novel.BaseSourceGateway') as source_gateway:
            source_gateway.use_cookies_from_browser.side_effect = CookieBrowserNotSupportedException('chrome')
            with self.assertRaises(SystemExit):
                novel_helper.set_cookies(source_gateway, 'chrome')

            source_gateway.use_cookies_from_browser.assert_called_with('chrome')

        with patch('novelsave.cli.helpers.novel.BaseSourceGateway') as source_gateway:
            novel_helper.set_cookies(source_gateway, 'chrome')

    @patch('novelsave.cli.helpers.novel.get_source_gateway')
    @patch('novelsave.services.source.SourceGateway')
    @patch('novelsave.services.NovelService')
    def test_create_novel_with_source(self, novel_service, source_gateway, get_source_gateway):
        novel_service.insert_novel.return_value = Novel()
        get_source_gateway.return_value = source_gateway
        source_gateway.novel_by_url.return_value = 'novel', ['chapters'], ['metadata']

        novel = novel_helper.create_novel('https://www.test.site', None, novel_service)

        novel_service.insert_novel.assert_called_with('novel')
        novel_service.insert_chapters.assert_called_with(novel_service.insert_novel.return_value, ['chapters'])
        novel_service.insert_metadata.assert_called_with(novel_service.insert_novel.return_value, ['metadata'])

        self.assertEqual(novel_service.insert_novel.return_value, novel)

    @patch('novelsave.services.NovelService')
    def test_get_novel_url(self, novel_service):
        novel_service.get_novel_by_url.return_value = None
        with self.assertRaises(ValueError):
            novel_helper.get_novel('https://test.site', novel_service)
        novel_service.get_novel_by_url.assert_called_with('https://test.site')

        novel = MagicMock()
        novel_service.get_novel_by_url.return_value = novel
        result = novel_helper.get_novel('https://test.com', novel_service)
        novel_service.get_novel_by_url.assert_called_with('https://test.com')
        self.assertEqual(novel, result)

    @patch('novelsave.services.NovelService')
    def test_get_novel_id(self, novel_service):
        novel_service.get_novel_by_id.return_value = None
        with self.assertRaises(ValueError):
            novel_helper.get_novel('1', novel_service)
        novel_service.get_novel_by_id.assert_called_with(1)

        novel = MagicMock()
        novel_service.get_novel_by_id.return_value = novel
        result = novel_helper.get_novel('2', novel_service)
        novel_service.get_novel_by_id.assert_called_with(2)
        self.assertEqual(novel, result)

    @patch('novelsave.services.NovelService')
    def test_get_novel_no_int_id(self, novel_service):
        with self.assertRaises(SystemExit):
            novel_helper.get_novel('h', novel_service)

    @patch('novelsave.cli.helpers.novel.get_novel', return_value=None)
    @patch('novelsave.cli.helpers.novel.create_novel', return_value='novel')
    def test_get_or_create_no_novel(self, create_novel, get_novel):
        result = novel_helper.get_or_create_novel('https://test.site')
        self.assertEqual('novel', result)

        with self.assertRaises(SystemExit):
            novel_helper.get_or_create_novel('1')

        with self.assertRaises(SystemExit):
            novel_helper.get_or_create_novel('something')

    @patch('novelsave.cli.helpers.novel.get_novel', return_value='novel')
    @patch('novelsave.cli.helpers.novel.create_novel', return_value='novel')
    def test_get_or_create_no_novel(self, create_novel, get_novel):
        result = novel_helper.get_or_create_novel('https://test.site')
        self.assertEqual('novel', result)

        result = novel_helper.get_or_create_novel('1')
        self.assertEqual('novel', result)


from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from ..base import Base


class NovelUrl(Base):
    __tablename__ = 'novel_urls'

    id = Column(Integer, primary_key=True)
    url = Column(String, nullable=False)

    novel_id = Column(Integer, ForeignKey('novels.id'))
    novel = relationship('Novel', back_populates='urls', lazy='joined')



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

from novelsave.core.entities.novel import Novel, Asset


class BasePathService(ABC):

    @abstractmethod
    def divide(self, r_path: Path) -> Path:
        """add additional sub folder to the parent depending on the file type"""

    @abstractmethod
    def novel_save_path(self, novel: Novel) -> Path:
        """:return absolute path to novel save directory"""

    @abstractmethod
    def novel_data_path(self, novel: Novel) -> Path:
        """:return absolute path to novel data directory"""

    @abstractmethod
    def asset_path(self, novel: Novel, asset: Asset) -> Path:
        """:return absolute path to the novel specific asset"""

    @abstractmethod
    def thumbnail_path(self, novel: Novel) -> Path:
        """:return absolute path to novel thumbnail file"""

    @abstractmethod
    def resolve_data_path(self, r_path: Union[Path, str]) -> Path:
        """converts relative path to data to absolute path"""

    @abstractmethod
    def relative_to_novel_dir(self, path: Path) -> Path:
        """provide the relative path segment from novels directory"""

    @abstractmethod
    def relative_to_data_dir(self, path: Path) -> Path:
        """provide the relative path segment from data directory"""



import json
import mimetypes
from functools import lru_cache
from pathlib import Path
from typing import Tuple, List, Dict

import lxml.html
from ebooklib import epub
from loguru import logger
from lxml.html import builder as E

from novelsave.core.entities.novel import Novel, Chapter, MetaData, NovelUrl
from novelsave.core.services import BaseNovelService, BasePathService, BaseFileService, BaseAssetService
from novelsave.core.services.packagers import BasePackager


class EpubPackager(BasePackager):
    def __init__(
            self,
            novel_service: BaseNovelService,
            file_service: BaseFileService,
            path_service: BasePathService,
            asset_service: BaseAssetService,
    ):
        self.novel_service = novel_service
        self.file_service = file_service
        self.path_service = path_service
        self.asset_service = asset_service

    def keywords(self) -> Tuple[str]:
        return 'epub',

    def package(self, novel: Novel):
        urls = self.novel_service.get_urls(novel)
        volumes = self.novel_service.get_volumes_with_chapters(novel)
        chapter_count = len([c for cl in volumes.values() for c in cl])
        metadata = self.novel_service.get_metadata(novel)
        logger.debug(f"Preparing to package to epub (id={novel.id}, title='{novel.title}', volumes={len(volumes)}, "
                     f"chapters={chapter_count}, metadata={len(metadata)})")

        book = epub.EpubBook()
        book.set_identifier(str(novel.id))
        book.set_title(novel.title)
        book.set_language(novel.lang)
        book.add_author(novel.author)
        self.set_cover(book, novel)

        if novel.synopsis:
            book.add_metadata('DC', 'description', novel.synopsis)
            logger.debug(f"Bound attributes to epub (synopsis={novel.synopsis[:min(30, len(novel.synopsis))]}…)")
        else:
            logger.debug(f"Bound attributes to epub (synopsis=None")

        for data in metadata:
            book.add_metadata(data.namespace, data.name, data.value, json.loads(data.others))
        logger.debug(f"Bound attributes to epub (metadata.count={len(metadata)})")

        book_preface = self.preface_html(novel, urls, metadata)
        book.add_item(book_preface)

        book_chapters = {}
        for volume, chapters in volumes.items():
            volume_tuple = (volume.index, volume.name)
            book_chapters[volume_tuple] = []

            for chapter in sorted(chapters, key=lambda c: c.index):
                epub_chapter = self.chapter_html(novel, chapter)
                book.add_item(epub_chapter)
                book_chapters[volume_tuple].append(epub_chapter)

        logger.debug(f"Added pages to epub (pages=(preface, chapters), count={chapter_count + 1}).")

        self.add_assets(book, novel)

        # table of contents
        book.toc = [book_preface]
        if len(book_chapters.keys()) == 1:  # no volume sections
            book.toc += list(book_chapters.values())[0]
            logger.debug(f"Built table of content of epub (type='single')")
        else:
            book.toc += [
                (epub.Section(volume[1]), tuple(book_chapters[volume]))
                for volume in sorted(book_chapters.keys(), key=lambda k: k[0])
            ]
            logger.debug(f"Built table of content of epub (type='multi')")

        # add default NCX and Nav file
        book.add_item(epub.EpubNcx())
        book.add_item(epub.EpubNav())

        book.spine = [book_preface] + [c for volume in book_chapters.values() for c in volume]

        path = self.destination(novel)
        path.parent.mkdir(parents=True, exist_ok=True)

        epub.write_epub(path, book, {})
        logger.debug(f"Saved epub file (path='{self.path_service.relative_to_novel_dir(path)}')")

        return path

    @lru_cache(maxsize=1)
    def destination(self, novel: Novel):
        path = self.path_service.novel_save_path(novel)
        return path / (path.name + '.epub')

    def set_cover(self, book: epub.EpubBook, novel: Novel):
        cover = None
        if novel.thumbnail_path is not None:
            possible_cover = self.path_service.resolve_data_path(Path(novel.thumbnail_path))
            if possible_cover.exists() and possible_cover.is_file():
                cover = possible_cover

        if cover is None:
            logger.debug(f"Copying cover image aborted (path='{novel.thumbnail_path}', reason='file does not exist')")
            return

        book.set_cover(cover.name, self.file_service.read_bytes(cover))
        logger.debug(f"Copied cover image to epub (path='{novel.thumbnail_path}')")

    @lru_cache(1)
    def path_mapping(self, novel: Novel) -> Dict[int, str]:
        assets = self.asset_service.downloaded_assets(novel)

        return {
            asset.id: f'../assets/{Path(asset.path).name}'
            for asset in assets
        }

    @lru_cache(1)
    def mapping_dict(self, novel: Novel):
        return self.asset_service.mapping_dict(self.path_mapping(novel))

    def chapter_html(self, novel: Novel, chapter: Chapter) -> epub.EpubHtml:
        content = self.asset_service.inject_assets(chapter.content, self.mapping_dict(novel))
        content_with_heading = f'<h1>{chapter.title}</h1>{content}'
        file_name = f'chapters/{str(chapter.index).zfill(4)}.xhtml'

        return epub.EpubHtml(
            title=chapter.title,
            content=content_with_heading,
            file_name=file_name,
            lang=novel.lang,
        )

    def add_assets(self, book: epub.EpubBook, novel: Novel):
        path_mapping = self.path_mapping(novel)

        assets = self.asset_service.downloaded_assets(novel)
        for asset in assets:
            file = self.path_service.resolve_data_path(asset.path)
            content = self.file_service.read_bytes(file)
            mimetype, encoding = mimetypes.guess_type(asset.path)

            image = epub.EpubItem(
                file_name=path_mapping[asset.id].lstrip('./'),
                media_type=mimetype,
                content=content,
            )

            book.add_item(image)

        logger.debug(f'Added assets to epub (count={len(assets)})')

    @staticmethod
    def metadata_display_value(data: MetaData):
        try:
            others = json.loads(data.others)
        except json.JSONDecodeError:
            others = {}

        others = [f'{key}={value}' for key, value in others.items()]
        postfix = f'(' + ', '.join(others) + ')' if others else ''

        return f'{data.value}{postfix}'

    def preface_html(self, novel: Novel, urls: List[NovelUrl], metadata: List[MetaData]) -> epub.EpubHtml:
        synopsis_section = E.DIV(
            E.H4('Synopsis'),
            E.DIV(
                *[E.P(para) for para in novel.synopsis.splitlines()],
                style="padding: 0 1rem",
            )
        )

        meta_by_name: Dict[str, List[MetaData]] = {}
        for item in metadata:
            meta_by_name.setdefault(item.name, []).append(item)

        metadata_sections = []
        for name, items in meta_by_name.items():
            item_strings = [self.metadata_display_value(item) for item in items]

            section = E.DIV(
                E.H4(name.capitalize()),
                E.DIV(
                    ', '.join(item_strings),
                    style="padding: 0 1rem",
                ),
            )

            metadata_sections.append(section)

        html = E.DIV(
            E.H1(novel.title, style="margin-bottom: 0;"),
            E.DIV(f'by: {novel.author}'),
            synopsis_section,
            *metadata_sections,
            E.DIV(
                E.H4('Sources'),
                E.DIV(
                    *[E.A(n_url.url, href=n_url.url) for n_url in urls],
                    style="padding: 0 1rem",
                )
            )
        )

        return epub.EpubHtml(
            title='Preface',
            file_name=f'preface.xhtml',
            content=lxml.html.tostring(html),
            lang=novel.lang,
        )



import unittest
from unittest.mock import patch

from loguru import logger

from novelsave.cli.controllers import package


@patch('novelsave.services.packagers.PackagerProvider')
@patch('novelsave.cli.controllers._package.get_novel')
class TestCompileController(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logger.remove()

    def test_compile_no_novel(self, get_novel, compiler_provider):
        get_novel.side_effect = ValueError()

        with self.assertRaises(SystemExit):
            package('https://novel.site', compiler_provider)

    def test_compile_with_novel(self, get_novel, compiler_provider):
        get_novel.return_value = 'novel'

        package('https://novel.site', compiler_provider)



from sqlalchemy import Column, Integer, String, ForeignKey, TIMESTAMP, func
from sqlalchemy.orm import relationship

from ..base import Base


class MetaData(Base):
    __tablename__ = 'novel_metadata'

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    value = Column(String, nullable=False)
    namespace = Column(String)
    others = Column(String)

    novel_id = Column(Integer, ForeignKey('novels.id'))
    novel = relationship('Novel', back_populates='novel_metadata')



from dataclasses import dataclass, field
from typing import List, Optional

from .chapter_dto import ChapterDTO


@dataclass
class VolumeDTO:
    id: Optional[int]
    index: int
    name: str

    chapters: List[ChapterDTO] = field(default_factory=lambda: [])



import functools
import sys

import click
from loguru import logger
from tqdm import tqdm

from . import controllers, helpers, groups, events
from ..containers import Application
from ..exceptions import NSError
from ..infrastructure.migrations import commands as migration_commands
from ..settings import config, DATABASE_URL, LOGGER_CONFIG
from ..utils.helpers import config_helper


def inject_dependencies():
    application = Application()
    application.config.from_dict(config)

    try:
        application.config.from_dict(config_helper.from_file())
    except FileNotFoundError:
        pass

    application.wire(modules=[events], packages=[controllers, helpers, groups])


def update_database_schema():
    migration_commands.migrate(DATABASE_URL)


@click.group()
@click.option('-d', '--debug', is_flag=True, help="Print debugging information to console")
@click.option('-p', '--plain', is_flag=True, help="Disable reactive and interactive elements")
def cli(debug: bool, plain: bool):
    logger_config = LOGGER_CONFIG.copy()
    if debug:
        logger_config['handlers'][0]['level'] = 'DEBUG'
    if plain:
        tqdm.__init__ = functools.partialmethod(tqdm.__init__, disable=True)

    logger.configure(**logger_config)

    update_database_schema()
    inject_dependencies()


@logger.catch()
def main():
    try:
        cli()
    except NSError as e:
        logger.exception(e)
        sys.exit(1)



import unittest
from datetime import datetime

from novelsave_sources import models as sm
from novelsave.core import dtos
from novelsave.utils.adapters import SourceAdapter


class TestSourceAdapter(unittest.TestCase):

    source_adapter = SourceAdapter()

    def test_novel_to_internal(self):
        test_novel = sm.Novel(
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            lang='language',
            url='link',
        )

        expected_novel = dtos.NovelDTO(
            id=None,
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            thumbnail_path=None,
            lang='language',
            url='link',
            last_updated=None,
        )

        actual_novel = self.source_adapter.novel_to_internal(test_novel)
        self.assertEqual(expected_novel, actual_novel)

    def test_novel_from_internal(self):
        test_novel = dtos.NovelDTO(
            id=1,
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            thumbnail_path='source/novel/assets/image.png',
            lang='language',
            url='link',
            last_updated=datetime.now(),
        )

        expected_novel = sm.Novel(
            title='title',
            author='author',
            synopsis='a nice description',
            thumbnail_url='thumbnail',
            lang='language',
            url='link',
        )

        actual_novel = self.source_adapter.novel_to_external(test_novel)
        self.assertEqual(expected_novel, actual_novel)

    def test_chapter_to_internal(self):
        test_chapter = sm.Chapter(
            index=1,
            no=20,
            title="title",
            paragraphs="paragraphs this is",
            volume=(1, "volume 1"),
            url="https://",
        )

        expected_chapter = dtos.ChapterDTO(
            index=1,
            title="title",
            content="paragraphs this is",
            volume=(1, "volume 1"),
            url="https://",
        )

        actual_chapter = self.source_adapter.chapter_to_internal(test_chapter)
        self.assertEqual(expected_chapter, actual_chapter)

    def test_chapter_from_internal(self):
        test_chapter = dtos.ChapterDTO(
            index=1,
            title="title",
            content="paragraphs this is",
            volume=(1, "volume 1"),
            url="https://",
        )

        expected_chapter = sm.Chapter(
            index=1,
            no=20,
            title="title",
            paragraphs="paragraphs this is",
            volume=(1, "volume 1"),
            url="https://",
        )

        actual_chapter = self.source_adapter.chapter_to_external(test_chapter)
        self.assertEqual(expected_chapter, actual_chapter)

    def test_chapter_content_to_internal(self):
        test_chapter = sm.Chapter(
            index=1,
            no=20,
            title="title",
            paragraphs="paragraphs this is",
            volume=(1, "volume 1"),
            url="https://",
        )

        expected_chapter = dtos.ChapterDTO(
            index=-1,
            title="",
            url="",
        )
        self.assertIsNone(expected_chapter.content)

        self.source_adapter.chapter_content_to_internal(test_chapter, expected_chapter)
        self.assertEqual(test_chapter.paragraphs, expected_chapter.content)

    def test_metadata_to_internal(self):
        test_metadata = sm.Metadata(
            name="name",
            value="value",
            others={"role": "something"},
            namespace="DC",
        )

        expected_metadata = dtos.MetaDataDTO(
            name="name",
            value="value",
            others={"role": "something"},
            namespace="DC",
        )

        actual_metadata = self.source_adapter.metadata_to_internal(test_metadata)
        self.assertEqual(expected_metadata, actual_metadata)

    def test_metadata_from_internal(self):
        test_metadata = dtos.MetaDataDTO(
            name="name",
            value="value",
            others={"role": "something"},
            namespace="DC",
        )

        expected_metadata = sm.Metadata(
            name="name",
            value="value",
            others={"role": "something"},
            namespace="DC",
        )

        actual_metadata = self.source_adapter.metadata_to_external(test_metadata)
        self.assertEqual(expected_metadata, actual_metadata)



from dataclasses import dataclass

from novelsave.core.dtos import ChapterDTO


class NSException(Exception):
    """base novelsave exception; can be caught and application may proceed once handled"""


class NSError(Exception):
    """base novelsave error; application must not continue once error is thrown"""


@dataclass
class CookieBrowserNotSupportedException(NSException):
    """the specified browser does not support cookie extraction"""
    browser: str


@dataclass
class ContentUpdateFailedException(NSException):
    chapter: ChapterDTO
    exception: Exception


@dataclass
class SourceNotFoundException(NSException):
    """source for the url was not found"""
    url: str

